# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# cython: language_level=3

import cython

from cython.view cimport array as cvarray

import numpy as np

cimport numpy as np

import cv2
from PIL import Image
from PIL.Image import Resampling

np.import_array()


cdef struct PixelRGBA:
    unsigned char r
    unsigned char g
    unsigned char b
    unsigned char a


cdef struct ImageInfo:
    int width
    int height
    PixelRGBA** img_ptr


cdef ImageInfo parse_img_info(image: Image):
    cdef ImageInfo info
    cdef unsigned long long ptr_val

    info.width = image.size[0]
    info.height = image.size[1]

    ptr_val = dict(image.getdata().unsafe_ptrs)['image']
    info.img_ptr = (<PixelRGBA**>ptr_val)

    return info


cdef inline int L24(PixelRGBA rgb):
    return rgb.r * 19595 + rgb.g * 38470 + rgb.b * 7471 + 0x8000


cdef inline unsigned char clip(float v):
    if v < 0.0:
        return 0
    if v >= 255.0:
        return 255

    return <unsigned char>v


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _c_lut(image: Image, int[:] lut):
    cdef ImageInfo info
    info = parse_img_info(image)

    for y in range(info.height):
        for x in range(info.width):
            info.img_ptr[y][x].r = lut[info.img_ptr[y][x].r]
            info.img_ptr[y][x].g = lut[info.img_ptr[y][x].g + 256]
            info.img_ptr[y][x].b = lut[info.img_ptr[y][x].b + 512]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int[:] c_histogram(image: Image):
    cdef ImageInfo info
    cdef int x, y
    cdef int[:] hist = cvarray(shape=(768,), itemsize=sizeof(int), format="i")

    info = parse_img_info(image)

    for x in range(768):
        hist[x] = 0

    for y in range(info.height):
        for x in range(info.width):
            hist[info.img_ptr[y][x].r] += 1
            hist[info.img_ptr[y][x].g + 256] += 1
            hist[info.img_ptr[y][x].b + 512] += 1

    return hist


def histogram(image: Image):
    cdef int[:] hist = c_histogram(image)
    cdef int i
    cdef int return_vals[768]

    for i in range(768):
        return_vals[i] = hist[i]

    return return_vals


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def autocontrast(image: Image, cutoff=0, ignore=None):
    if image.mode != "RGB":
        image = image.convert("RGB")
    cdef int layer = 0
    cdef int* h
    cdef int i, lo, hi, ix, cut, n
    cdef double scale, offset
    cdef int[:] histogram
    cdef int[:] lut = cvarray(shape=(768,), itemsize=sizeof(int), format="i")

    histogram = c_histogram(image)

    for layer in range(0, 768, 256):
        h = &histogram[layer]

        if ignore is not None:
            # get rid of outliers
            try:
                h[ignore] = 0
            except TypeError:
                # assume sequence
                for ix in ignore:
                    h[ix] = 0
        if cutoff:
            # cut off pixels from both ends of the histogram
            if not isinstance(cutoff, tuple):
                cutoff = (cutoff, cutoff)
            # get number of pixels
            n = 0
            for ix in range(256):
                n = n + h[ix]
            # remove cutoff% pixels from the low end
            cut = n * cutoff[0] // 100
            for lo in range(256):
                if cut > h[lo]:
                    cut = cut - h[lo]
                    h[lo] = 0
                else:
                    h[lo] -= cut
                    cut = 0
                if cut <= 0:
                    break
            # remove cutoff% samples from the high end
            cut = n * cutoff[1] // 100
            for hi in range(255, -1, -1):
                if cut > h[hi]:
                    cut = cut - h[hi]
                    h[hi] = 0
                else:
                    h[hi] -= cut
                    cut = 0
                if cut <= 0:
                    break
        # find lowest/highest samples after preprocessing
        for lo in range(256):
            if h[lo]:
                break
        for hi in range(255, -1, -1):
            if h[hi]:
                break
        if hi <= lo:
            # don't bother
            for i in range(256):
                lut[layer + i] = i
        else:
            scale = 255.0 / (hi - lo)
            offset = -lo * scale
            for ix in range(256):
                i = ix
                ix = (int)(ix * scale + offset)
                if ix < 0:
                    ix = 0
                elif ix > 255:
                    ix = 255
                lut[layer + i] = ix

        layer += 256

    _c_lut(image, lut)

    return image


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def equalize(image: Image):
    if image.mode != "RGB":
        image = image.convert("RGB")
    cdef int[:] h
    cdef int[:] lut = cvarray(shape=(768,), itemsize=sizeof(int), format="i")

    h = c_histogram(image)

    cdef int b, histo_len, histo_sum, i, n, step, num
    cdef int histo[256]

    for b in range(0, 768, 256):
        histo_len = 0
        histo_sum = 0

        for i in range(256):
            num = h[b + i]
            if num > 0:
                histo[histo_len] = num
                histo_sum += num
                histo_len += 1

        if histo_len <= 1:
            for i in range(256):
                lut[b + i] = i
        else:
            step = (histo_sum - histo[histo_len - 1]) // 255
            if not step:
                for i in range(256):
                    lut[b + i] = i
            else:
                n = step // 2
                for i in range(256):
                    lut[b + i] = n // step
                    n = n + h[i + b]

    _c_lut(image, lut)

    return image


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def posterize(image: Image, bits: int):
    if image.mode != "RGB":
        image = image.convert("RGB")

    cdef int[:] lut = cvarray(shape=(768,), itemsize=sizeof(int), format="i")
    cdef int i, b, c_bits
    cdef unsigned char mask

    c_bits = bits

    mask = ~(2 ** (8 - c_bits) - 1)
    for b in range(0, 768, 256):
        for i in range(256):
            lut[b + i] = i & mask
    _c_lut(image, lut)
    return image


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def solarize(image: Image, threshold: int = 128):
    if image.mode != "RGB":
        image = image.convert("RGB")

    cdef int[:] lut = cvarray(shape=(768,), itemsize=sizeof(int), format="i")
    cdef int i, b, c_threshold
    cdef ImageInfo info

    c_threshold = threshold

    for b in range(0, 768, 256):
        for i in range(256):
            if i < c_threshold:
                lut[b + i] = i
            else:
                lut[b + i] = 255 - i

    _c_lut(image, lut)
    return image


@cython.boundscheck(False)
@cython.wraparound(False)
def color(image: Image, factor: float):
    if image.mode != "RGB":
        image = image.convert("RGB")

    cdef ImageInfo info
    cdef int x, y
    cdef float grey_val, c_factor

    info = parse_img_info(image)
    c_factor = factor

    for y in range(info.height):
        for x in range(info.width):
            grey_val = <float>(L24(info.img_ptr[y][x]) >> 16)
            if 0.0 <= c_factor and c_factor <= 1.0:
                info.img_ptr[y][x].r = <unsigned char>(info.img_ptr[y][x].r * c_factor + grey_val * (1 - c_factor))
                info.img_ptr[y][x].g = <unsigned char>(info.img_ptr[y][x].g * c_factor + grey_val * (1 - c_factor))
                info.img_ptr[y][x].b = <unsigned char>(info.img_ptr[y][x].b * c_factor + grey_val * (1 - c_factor))
            else:
                info.img_ptr[y][x].r = clip(info.img_ptr[y][x].r * c_factor + grey_val * (1 - c_factor))
                info.img_ptr[y][x].g = clip(info.img_ptr[y][x].g * c_factor + grey_val * (1 - c_factor))
                info.img_ptr[y][x].b = clip(info.img_ptr[y][x].b * c_factor + grey_val * (1 - c_factor))

    return image


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def contrast(image: Image, factor: float):
    if image.mode != "RGB":
        image = image.convert("RGB")

    cdef ImageInfo info
    cdef int i_sum, x, y
    cdef float c_factor, f_mean

    info = parse_img_info(image)
    c_factor = factor

    f_mean = 0
    i_sum = 0
    for y in range(info.height):
        for x in range(info.width):
            i_sum += L24(info.img_ptr[y][x]) >> 16
    f_mean = i_sum / (info.height * info.width)

    for y in range(info.height):
        for x in range(info.width):
            if 0.0 <= c_factor and c_factor <= 1.0:
                info.img_ptr[y][x].r = <unsigned char>(info.img_ptr[y][x].r * c_factor + f_mean * (1 - c_factor))
                info.img_ptr[y][x].g = <unsigned char>(info.img_ptr[y][x].g * c_factor + f_mean * (1 - c_factor))
                info.img_ptr[y][x].b = <unsigned char>(info.img_ptr[y][x].b * c_factor + f_mean * (1 - c_factor))
            else:
                info.img_ptr[y][x].r = clip(info.img_ptr[y][x].r * c_factor + f_mean * (1 - c_factor))
                info.img_ptr[y][x].g = clip(info.img_ptr[y][x].g * c_factor + f_mean * (1 - c_factor))
                info.img_ptr[y][x].b = clip(info.img_ptr[y][x].b * c_factor + f_mean * (1 - c_factor))

    return image


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def brightness(image: Image, factor: float):
    if image.mode != "RGB":
        image = image.convert("RGB")

    cdef ImageInfo info
    cdef int x, y
    cdef float c_factor

    info = parse_img_info(image)
    c_factor = factor

    for y in range(info.height):
        for x in range(info.width):
            if 0.0 <= c_factor and c_factor <= 1.0:
                info.img_ptr[y][x].r = <unsigned char>(info.img_ptr[y][x].r * c_factor)
                info.img_ptr[y][x].g = <unsigned char>(info.img_ptr[y][x].g * c_factor)
                info.img_ptr[y][x].b = <unsigned char>(info.img_ptr[y][x].b * c_factor)
            else:
                info.img_ptr[y][x].r = clip(info.img_ptr[y][x].r * c_factor)
                info.img_ptr[y][x].g = clip(info.img_ptr[y][x].g * c_factor)
                info.img_ptr[y][x].b = clip(info.img_ptr[y][x].b * c_factor)

    return image


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def sharpness(image: Image, factor: float):
    if image.mode != "RGB":
        image = image.convert("RGB")

    cdef ImageInfo info
    cdef int x, y, i, j
    cdef float c_factor
    cdef float smooth_kernel[3][3]
    smooth_kernel[0][:] = [1 / 13., 1 / 13., 1 / 13.]
    smooth_kernel[1][:] = [1 / 13., 5 / 13., 1 / 13.]
    smooth_kernel[2][:] = [1 / 13., 1 / 13., 1 / 13.]
    cdef float r, g, b, div

    info = parse_img_info(image)
    c_factor = factor

    for i in range(3):
        for j in range(3):
            smooth_kernel[i][j] = smooth_kernel[i][j] * (1 - c_factor)

    smooth_kernel[1][1] += c_factor

    for y in range(1, info.height - 1):
        for x in range(1, info.width - 1):
            r = g = b = div = 0

            for i in range(3):
                for j in range(3):
                    r += smooth_kernel[i][j] * info.img_ptr[y + i - 1][x + j - 1].r
                    g += smooth_kernel[i][j] * info.img_ptr[y + i - 1][x + j - 1].g
                    b += smooth_kernel[i][j] * info.img_ptr[y + i - 1][x + j - 1].b

            info.img_ptr[y][x].r = clip(r)
            info.img_ptr[y][x].g = clip(g)
            info.img_ptr[y][x].b = clip(b)

    return image


def _convert_flag_pil_to_cv(flag: Resampling) -> int:
    flag_map = {
        Resampling.NEAREST: cv2.INTER_NEAREST,
        Resampling.BOX: cv2.INTER_NEAREST,
        Resampling.BILINEAR: cv2.INTER_LINEAR,
        Resampling.HAMMING: cv2.INTER_LINEAR,
        Resampling.BICUBIC: cv2.INTER_CUBIC,
        Resampling.LANCZOS: cv2.INTER_LANCZOS4
    }

    if flag in flag_map:
        return flag_map[flag]
    return cv2.INTER_LINEAR


def rotate(image: Image, angle: float, resample: Resampling = Resampling.BILINEAR):
    if image.mode != "RGB":
        image = image.convert("RGB")

    image = np.asarray(image)
    image_center = tuple(np.array(image.shape[1::-1]) / 2)
    rot_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
    flags = _convert_flag_pil_to_cv(resample)
    result = cv2.warpAffine(image, rot_mat, image.shape[1::-1], flags=flags)
    result = Image.fromarray(result)
    return result


def translate_x_rel(image: Image, pct: float, resample: Resampling = Resampling.BILINEAR):
    if image.mode != "RGB":
        image = image.convert("RGB")

    pixels = pct * image.width
    image = np.asarray(image)
    aff_mat = np.asarray((1, 0, -pixels, 0, 1, 1)).reshape([2, 3])
    flags = _convert_flag_pil_to_cv(resample)
    result = cv2.warpAffine(image, aff_mat, image.shape[1::-1], flags=flags)
    return Image.fromarray(result)


def translate_y_rel(image: Image, pct: float, resample: Resampling = Resampling.BILINEAR):
    if image.mode != "RGB":
        image = image.convert("RGB")

    pixels = pct * image.height
    image = np.asarray(image)
    aff_mat = np.asarray((1, 0, 0, 0, 1, -pixels)).reshape([2, 3])
    flags = _convert_flag_pil_to_cv(resample)
    result = cv2.warpAffine(image, aff_mat, image.shape[1::-1], flags=flags)
    return Image.fromarray(result)


def shear_x(image: Image, factor: float, resample: Resampling = Resampling.BILINEAR):
    if image.mode != "RGB":
        image = image.convert("RGB")

    image = np.asarray(image)
    aff_mat = np.asarray((1, -factor, 0, 0, 1, 0)).reshape([2, 3])
    flags = _convert_flag_pil_to_cv(resample)
    result = cv2.warpAffine(image, aff_mat, image.shape[1::-1], flags=flags)
    return Image.fromarray(result)


def shear_y(image: Image, factor: float, resample: Resampling = Resampling.BILINEAR):
    if image.mode != "RGB":
        image = image.convert("RGB")

    image = np.asarray(image)
    aff_mat = np.asarray((1, 0, 0, -factor, 1, 0)).reshape([2, 3])
    flags = _convert_flag_pil_to_cv(resample)
    result = cv2.warpAffine(image, aff_mat, image.shape[1::-1], flags=flags)
    return Image.fromarray(result)


@cython.boundscheck(False)
@cython.wraparound(False)
def blend(img_to_mix: Image, np.ndarray[np.float32_t, ndim=3] img_dst, weight: float) -> None:
    if img_to_mix.mode != "RGB":
        img_to_mix = img_to_mix.convert("RGB")

    cdef ImageInfo info = parse_img_info(img_to_mix)
    cdef float c_weight = weight

    for y in range(info.height):
        for x in range(info.width):
            img_dst[y, x, 0] += c_weight * info.img_ptr[y][x].r
            img_dst[y, x, 1] += c_weight * info.img_ptr[y][x].g
            img_dst[y, x, 2] += c_weight * info.img_ptr[y][x].b



"""
Config Helpers for OTX Training
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Tuple

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.helper import create
from otx.api.entities.model_template import ModelTemplate, parse_model_template


def get_config_and_task_name(template_file_path: str) -> Tuple[ConfigurableParameters, str]:
    """Return configurable parameters and model name given template path

    Args:
        template_file_path (str): template path

    Returns:
        Tuple[ConfigurableParameters, str]: Configurable parameters, model name
    """
    model_template: ModelTemplate = parse_model_template(template_file_path)
    hyper_parameters: dict = model_template.hyper_parameters.data
    config: ConfigurableParameters = create(hyper_parameters)
    return config, model_template.name



"""Torchvision Model Backbone Class Generation.

For torchvision backbone support for OTX models,
this is a code that converts torchvision backbone classes to match
the mmcv backbone class format and registers them in the mmcv registry.
This copied the format of "mmdet/models/backbones/imgclsmob.py"
as it is and made some modifications & code cleaning.
"""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import inspect
import os

from mmcv.cnn import build_activation_layer, build_norm_layer
from torch import nn
from torch.hub import load_state_dict_from_url
from torch.nn.modules.batchnorm import _BatchNorm
from torchvision import models

from ..builder import TORCHVISION_BACKBONES

# pylint: disable=protected-access, assignment-from-no-return, no-value-for-parameter, too-many-statements


def get_torchvision_models():
    """Get torchvision backbones of current version."""
    torchvision_urls = {}
    torchvision_models = {}
    for model_key, model_value in models.__dict__.items():
        if callable(model_value) and model_key[0].islower() and model_key[0] != "_":
            torchvision_models[model_key] = model_value
        elif inspect.ismodule(model_value) and hasattr(model_value, "model_urls"):
            torchvision_urls.update(model_value.model_urls)
    return torchvision_models, torchvision_urls


TORCHVISION_MODELS, TORCHVISION_MODEL_URLS = get_torchvision_models()


def replace_activation(model, activation_cfg):
    """Replace Activate function (copy from mmdet)."""
    for name, module in model._modules.items():
        if len(list(module.children())) > 0:
            model._modules[name] = replace_activation(module, activation_cfg)
        if name == "activ":
            if activation_cfg["type"] == "torch_swish":
                model._modules[name] = nn.SiLU()
            else:
                model._modules[name] = build_activation_layer(activation_cfg)
    return model


def replace_norm(model, cfg):
    """Replace Norm function (copy from mmdet)."""
    for name, module in model._modules.items():
        if len(list(module.children())) > 0:
            model._modules[name] = replace_norm(module, cfg)
        if name == "bn":
            model._modules[name] = build_norm_layer(cfg, num_features=module.num_features)[1]
    return model


def resnet_forward(self, x):
    """Resnet forward function for wrapping model (refer to torchvision)."""
    outputs = []
    y = x
    stages = [self.layer1, self.layer2, self.layer3, self.layer4]
    last_stage = max(self.out_indices)
    y = self.conv1(y)
    y = self.bn1(y)
    y = self.relu(y)
    y = self.maxpool(y)
    for i, stage in enumerate(stages):
        y = stage(y)
        if i in self.out_indices:
            outputs.append(y)
        if i == last_stage:
            break
    return tuple(outputs)


def shufflenet_forward(self, x):
    """Shufflenet forward function for wrapping model (refer to torchvision)."""
    outputs = []
    y = x
    y = self.conv1(y)
    y = self.maxpool(y)
    stages = [self.stage2, self.stage3, self.stage4, self.conv5]
    last_stage = max(self.out_indices)
    for i, stage in enumerate(stages):
        y = stage(y)
        if i in self.out_indices:
            outputs.append(y)
        if i == last_stage:
            break
    return tuple(outputs)


def multioutput_forward(self, x):
    """Multioutput forward function for new model (copy from mmdet)."""
    outputs = []
    y = x

    last_stage = max(self.out_indices)
    if hasattr(self, "features"):
        stages = self.features
    elif hasattr(self, "layers"):
        stages = self.layers
    else:
        raise ValueError(f"Not supported multioutput forward: {self}")

    for i, stage in enumerate(stages):
        y = stage(y)
        temp_s = str(i) + " " + str(y.shape)
        if i in self.out_indices:
            outputs.append(y)
            temp_s += "*"
        if self.verbose:
            print(temp_s)
        if i == last_stage:
            break
    return tuple(outputs)


def train(self, mode=True):
    """Train forward function for new model (copy from mmdet)."""
    super(self.__class__, self).train(mode)

    if hasattr(self, "features"):
        stages = self.features
    elif hasattr(self, "layers"):
        stages = self.layers
    else:
        raise ValueError(f"Not supported multioutput forward: {self}")

    for i in range(self.frozen_stages + 1):
        temp_m = stages[i]
        temp_m.eval()
        for param in temp_m.parameters():
            param.requires_grad = False

    if mode and self.norm_eval:
        for mmodule in self.modules():
            # trick: eval have effect on BatchNorm only
            if isinstance(mmodule, _BatchNorm):
                mmodule.eval()


def init_weights(self):
    """Init weights function for new model (copy from mmdet)."""
    if self.init_cfg.get("Pretrained", False) and self.model_urls:
        state_dict = load_state_dict_from_url(self.model_urls)
        self.load_state_dict(state_dict)


def generate_torchvision_backbones():
    """Regist Torchvision Backbone into mmX Registry (copy from mmdet)."""
    for model_name, model_builder in TORCHVISION_MODELS.items():

        def closure(model_name, model_builder):
            """Get Model builder for mmcv (copy from mmdet)."""

            class TorchvisionModelWrapper(nn.Module):  # pylint: disable=abstract-method
                """Torchvision Model to MMX.model Wrapper (copy from mmdet)."""

                def __init__(
                    self,
                    *args,
                    out_indices=(0, 1, 2, 3),
                    frozen_stages=0,
                    norm_eval=False,
                    verbose=False,
                    activation_cfg=None,
                    norm_cfg=None,
                    init_cfg=None,
                    **kwargs,
                ):
                    super().__init__()
                    models_cache_root = kwargs.get("root", os.path.join("~", ".torch", "models"))
                    model = model_builder(*args, **kwargs)
                    if activation_cfg:
                        model = replace_activation(model, activation_cfg)
                    if norm_cfg:
                        model = replace_norm(model, norm_cfg)
                    model.out_indices = out_indices
                    model.frozen_stages = frozen_stages
                    model.norm_eval = norm_eval
                    model.verbose = verbose
                    model_name = str(self).strip("()")
                    model.model_urls = (
                        TORCHVISION_MODEL_URLS[model_name] if model_name in TORCHVISION_MODEL_URLS else None
                    )
                    model.models_cache_root = models_cache_root
                    model.init_cfg = init_cfg
                    model.init_weights = init_weights.__get__(model)
                    if hasattr(model, "features") and isinstance(model.features, nn.Sequential):
                        # Save original forward, just in case.
                        model.forward_single_output = model.forward
                        model.forward = multioutput_forward.__get__(model)
                        model.train = train.__get__(model)

                        model.output = None
                        for i, _ in enumerate(model.features):
                            if out_indices is not None and i > max(out_indices):
                                model.features[i] = None
                    elif hasattr(model, "layers") and isinstance(model.layers, nn.Sequential):
                        # Save original forward, just in case.
                        model.forward_single_output = model.forward
                        model.forward = multioutput_forward.__get__(model)
                        model.train = train.__get__(model)

                        model.classifier = None
                        for i, _ in enumerate(model.layers):
                            if out_indices is not None and i > max(out_indices):
                                model.layers[i] = None
                    elif model_name.startswith(("resne", "wide_resne")):
                        # torchvision.resne* -> resnet_forward
                        model.forward = resnet_forward.__get__(model)
                        model.fc = None
                    elif model_name.startswith("shufflenet"):
                        model.forward = shufflenet_forward.__get__(model)
                        model.fc = None
                    else:
                        raise ValueError(
                            "Failed to automatically wrap backbone network. "
                            f"Object of type {model.__class__} has no valid attribute called "
                            '"features".'
                        )
                    self.__dict__.update(model.__dict__)

            TorchvisionModelWrapper.__name__ = model_name
            return TorchvisionModelWrapper

        TORCHVISION_BACKBONES.register_module(name=model_name, module=closure(model_name, model_builder))


generate_torchvision_backbones()



"""This module implements the Metric entities."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
import datetime
import logging
import math
from enum import Enum
from typing import Generic, List, Optional, Sequence, TypeVar, Union

import numpy as np

from otx.api.utils.time_utils import now


class MetricEntity(metaclass=abc.ABCMeta):
    """This interface represents a metric, which is the smallest building block for the performance statistics.

    It only contains the name of the metric.
    See also :class:`MetricsGroup` and :class:`Performance` for the structure of performance statistics.
    """

    __name = None

    @property
    def name(self):
        """Returns the name of the Metric Entity."""
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @staticmethod
    def type() -> str:
        """Returns the type of the MetricEntity, e.g. "curve."""


class CountMetric(MetricEntity):
    """This metric represents an integer value.

    Args:
        name: The name of the metric
        value: The value of the metric

    Example:
        The count for number of images in a project

    >>> count_metric = CountMetric(name="Number of images", value=20)

    """

    value: int

    def __init__(self, name: str, value: int):
        self.name = name
        self.value = value

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "count"


class InfoMetric(MetricEntity):
    """This metric represents a string value.

    Args:
        name: The name of the info metric
        value: The info of the metric

    Example:
        An info metric of training from scratch

    >>> info_metric = InfoMetric(name="Model info", value="This model is trained from scratch")

    """

    value: str

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "string"


class DateMetric(MetricEntity):
    """This metric represents a date time value.

    Args:
        name: The name of the date metric
        date: The datetime value of the metric

    Example:
        A DateMetric for model creation date (e.g., now).

    >>> metric = DateMetric(name="Model creation", date=datetime.datetime.now(datetime.timezone.utc))

    """

    date: datetime.datetime

    def __init__(self, name: str, date: Optional[datetime.datetime] = None):
        self.name = name
        if date is None:
            date = now()
        self.date = date

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "date"


class ScoreMetric(MetricEntity):
    """This metric represents a float value.

    This metric is typically used for storing performance metrics, such as accuracy, f-measure, dice score, etc.

    Args:
        name: The name of the score
        value: The value of the score

    Example:
        Accuracy of a model

    >>> score_metric = ScoreMetric(name="Model accuracy", value=0.5)

    """

    def __init__(self, name: str, value: float):
        self.name = name
        self.value = value

        if math.isnan(value):
            raise ValueError("The value of a ScoreMetric is not allowed to be NaN.")

    def __eq__(self, other: object) -> bool:
        """Returns True if the score metrics are equal."""
        if not isinstance(other, ScoreMetric):
            return False
        return self.name == other.name and self.value == other.value

    def __repr__(self):
        """Returns the representation of the score metric."""
        return f"ScoreMetric(name=`{self.name}`, score=`{self.value}`)"

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "score"


class DurationMetric(MetricEntity):
    """This metric represents a duration metric, which include hour (int), minute (int), and second (float).

    Args:
        name: The name of the duration metric
        hour: The hour value of the metric
        minute: The minute value of the metric
        second: The second value of the metric

    Example:
        Creating a metric for training duration of 1 hour 5 minutes.

    >>> duration_metric = DurationMetric(name="Training duration", hour=1, minute=5, second=0)

    """

    def __init__(self, name: str, hour: int, minute: int, second: float):
        self.hour = hour
        self.minute = minute
        self.second = second
        self.name = name

    def get_duration_string(self) -> str:
        """Returns the string representation of the duration.

        Example:
            Duration string of 1 hour 1 minute and 1.50 seconds.

        >>> from otx.api.entities.metrics import DurationMetric
        >>> dur_met = DurationMetric("test", 1, 1, 1.5)  # 1 hour 1 minute and 1.5 seconds
        >>> dur_met.get_duration_string()
        '1 hour 1 minute 1.50 seconds'

        Returns:
            the string representation of the duration.
        """
        output: str = ""
        if self.hour != 0:
            output += f"{self.hour} hour{'s ' if self.hour > 1 else ' '}"
        if self.minute != 0:
            output += f"{self.minute} minute{'s ' if self.minute > 1 else ' '}"
        if self.second != 0:
            output += f"{self.second:.02f} second{'s ' if self.second > 1 else ' '}"
        output = output.strip()
        return output

    @staticmethod
    def from_seconds(name: str, seconds: float) -> "DurationMetric":
        """Returns a duration metrics, with name and converted durations from seconds.

        Example:
            Converting 70 seconds to duration metric.

        >>> from otx.api.entities.metrics import DurationMetric
        >>> dur_met = DurationMetric.from_seconds("test", 70)  # 1 hour 1 minute and 1.5 seconds
        >>> dur_met.get_duration_string()
        '1 minute 10.00 seconds'

        Args:
            name
            seconds

        Returns:
            DurationMetric: the duration metric with name and converted durations from seconds.
        """
        hour = int(seconds // 3600)
        modulo = seconds % 3600
        minute = int(modulo // 60)
        second = modulo % 60
        return DurationMetric(name=name, hour=hour, minute=minute, second=second)

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "duration"


class CurveMetric(MetricEntity):
    """This metric represents a curve. The coordinates are represented as x and y lists.

    Example:
        A line curve of: [(0,1), (1, 5), (2, 8)]

    >>> CurveMetric("Line", xs=[0, 1, 2], ys=[1, 5, 8])
    CurveMetric(name=`Line`, ys=(3 values), xs=(3 values))

    A curve can also be defined only using the y values. For example, a loss curve of loss values: [0.5, 0.2, 0.1].
    The x values will be automatically generated as a 1-based index (1, 2, 3, ...)

    >>> CurveMetric("Loss", ys=[0.5, 0.2, 0.1])
    CurveMetric(name=`Loss`, ys=(3 values), xs=(None values))

    Args:
        name: The name of the curve
        xs: the list of floats in x-axis
        ys: the list of floats in y-axis
    """

    def __init__(self, name: str, ys: List[float], xs: Optional[List[float]] = None):
        self.name = name
        self.__ys = ys
        if xs is not None:
            if len(xs) != len(self.__ys):
                raise ValueError(f"Curve error must contain the same length for x and y: ({len(xs)} vs {len(self.ys)})")
            self.__xs = xs
        else:
            # if x values are not provided, set them to the 1-index of the y values
            self.__xs = list(range(1, len(self.__ys) + 1))

    @property
    def ys(self) -> List[float]:
        """Returns the list of floats on y-axis."""
        return self.__ys

    @property
    def xs(self) -> List[float]:
        """Returns the list of floats on x-axis."""
        return self.__xs

    def __repr__(self):
        """Returns the string representation of the object."""
        return (
            f"CurveMetric(name=`{self.name}`, ys=({len(self.ys)} values), "
            f"xs=({len(self.xs) if self.xs is not None else 'None'} values))"
        )

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "curve"


class MatrixMetric(MetricEntity):
    """This metric represents a matrix. The cells are represented as a list of lists of integers.

    In the case of a confusion matrix, the rows represent the ground truth items and the columns represent the
    predicted items.

    Example:
        A matrix of: [[4,0,1], [0,3,2], [1,2,2]]

        >>> MatrixMetric("Confusion Matrix", matrix_values=np.array([[4,0,1], [0,3,2], [1,2,2]]))
        MatrixMetric(name=`Confusion Matrix`, matrix_values=(3x3) matrix, row labels=None, column labels=None)

    Args:
        name: The name of the matrix
        matrix_values: the matrix data
        row_labels: labels for the rows
        column_labels: labels for the columns
        normalize: set to True to normalize each row of the matrix
    """

    __row_labels: Optional[List[str]] = None
    __column_labels: Optional[List[str]] = None

    # pylint: disable=too-many-arguments; Requires refactor
    def __init__(
        self,
        name: str,
        matrix_values: np.ndarray,
        row_labels: Optional[List[str]] = None,
        column_labels: Optional[List[str]] = None,
        normalize: bool = False,
    ):
        self.name = name
        self.__matrix_values = matrix_values
        self.__matrix_values.astype(np.float32)

        if row_labels is not None:
            self.__row_labels = row_labels
            if self.__matrix_values.shape[0] != len(self.__row_labels):
                raise ValueError(
                    f"Number of rows of the matrix and number of row labels must be equal. The shape "
                    f"has {self.__matrix_values.shape[0]} rows and {len(self.__row_labels)} row labels"
                )

        if column_labels is not None:
            self.__column_labels = column_labels
            if self.__matrix_values.shape[1] != len(self.__column_labels):
                raise ValueError(
                    f"Number of columns of the matrix and number of column labels must be equal. The "
                    f"shape has {self.__matrix_values.shape[1]} columns and {len(self.__column_labels)} column "
                    "labels"
                )

        if normalize:
            self.normalize()

    @property
    def matrix_values(self) -> np.ndarray:
        """Returns the matrix data."""
        return self.__matrix_values

    @property
    def row_labels(self) -> Optional[List[str]]:
        """Returns the row labels."""
        return self.__row_labels

    @property
    def column_labels(self) -> Optional[List[str]]:
        """Returns the column labels."""
        return self.__column_labels

    def normalize(self):
        """Normalizes the confusion matrix by dividing by the sum of the rows."""
        self.__matrix_values = self.__matrix_values.astype(np.float32) / self.__matrix_values.astype(np.float32).sum(
            axis=1, keepdims=True
        )  # Divide all values by the sum of its row

        if not np.all(self.__matrix_values.sum(axis=1, keepdims=True) > 0):
            self.__matrix_values = np.nan_to_num(self.__matrix_values)

            logger = logging.getLogger(__name__)
            logger.warning("Replacing NaN in the matrix with zeroes since the sum of one (or more) row(s) was zero.")

    def __repr__(self):
        """Returns the string representation of the object."""
        return (
            f"MatrixMetric(name=`{self.name}`, matrix_values=({self.__matrix_values.shape[0]}x"
            f"{self.__matrix_values.shape[1]}) matrix, row labels={self.__row_labels}, column labels"
            f"={self.__column_labels})"
        )

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "matrix"


class NullMetric(MetricEntity):
    """Represents 'Metric not found'."""

    def __init__(self) -> None:
        self.name = "NullMetric"

    def __repr__(self):
        """Returns the string representation of the object."""
        return "NullMetric()"

    def __eq__(self, other):
        """Returns True if the other object is a NullMetric."""
        return isinstance(other, NullMetric)

    @staticmethod
    def type():
        """Returns the type of the MetricEntity."""
        return "null"


class VisualizationType(Enum):
    """This enum defines how the metrics will be visualized on the UI."""

    TEXT = 0
    RADIAL_BAR = 1
    BAR = 2
    LINE = 3
    MATRIX = 4


class ColorPalette(Enum):
    """Enum class specifying the color palette to be used by the UI to display statistics.

    If the statistics are per label, set to LABEL so the UI will use the label color palette.
    Otherwise, set to DEFAULT (allow the UI to choose a color palette)
    """

    DEFAULT = 0
    LABEL = 1


class VisualizationInfo:
    """This represents the visualization info a metrics group. See :class:`MetricsGroup`."""

    __type: VisualizationType
    name: str  # todo: this should be a part of MetricsGroup, not the visualization info.

    def __init__(
        self,
        name: str,
        visualisation_type: VisualizationType,
        palette: ColorPalette = ColorPalette.DEFAULT,
    ):
        self.__type = visualisation_type
        self.name = name
        self.palette = palette

    @property
    def type(self) -> VisualizationType:
        """Returns the type of the visualization."""
        return self.__type

    def __repr__(self):
        """Returns the string representation of the object."""
        return f"VisualizationInfo(name='{self.name}', type='{self.type.name}', palette='{self.palette.name}')"


class TextChartInfo(VisualizationInfo):
    """This represents a visualization using text, which uses only a single string."""

    def __init__(
        self,
        name: str,
    ):
        super().__init__(name, VisualizationType.TEXT)

    def __repr__(self):
        """Returns the string representation of the object."""
        return f"TextChartInfo(name='{self.name}, 'type='{self.type}')"


class LineChartInfo(VisualizationInfo):
    """This represents a visualization using a line chart."""

    x_axis_label: str
    y_axis_label: str

    def __init__(
        self,
        name: str,
        x_axis_label: str = None,
        y_axis_label: str = None,
        palette: ColorPalette = ColorPalette.DEFAULT,
    ):
        super().__init__(name, VisualizationType.LINE, palette)
        if x_axis_label is None:
            x_axis_label = ""
        if y_axis_label is None:
            y_axis_label = ""

        self.x_axis_label = x_axis_label
        self.y_axis_label = y_axis_label

    def __repr__(self):
        """Returns the string representation of the object."""
        return (
            f"LineChartInfo(name='{self.name}, 'type='{self.type}', x_axis_label='{self.x_axis_label}', "
            f"y_axis_label='{self.y_axis_label}')"
        )


class BarChartInfo(VisualizationInfo):
    """This represents a visualization using a bar chart."""

    def __init__(
        self,
        name: str,
        palette: ColorPalette = ColorPalette.DEFAULT,
        visualization_type: VisualizationType = VisualizationType.BAR,
    ):
        if visualization_type not in (
            VisualizationType.BAR,
            VisualizationType.RADIAL_BAR,
        ):
            raise ValueError("Visualization type for BarChartInfo must be BAR or RADIAL_BAR")
        super().__init__(name, visualization_type, palette)

    def __repr__(self):
        """Returns the string representation of the object."""
        return f"BarChartInfo(name='{self.name}', type='{self.type}')"


class MatrixChartInfo(VisualizationInfo):
    """This represents a visualization using a matrix."""

    header: str
    row_header: str
    column_header: str

    # pylint: disable=too-many-arguments; Requires refactor
    def __init__(
        self,
        name: str,
        header: str = None,
        row_header: str = None,
        column_header: str = None,
        palette: ColorPalette = ColorPalette.DEFAULT,
    ):
        super().__init__(name, VisualizationType.MATRIX, palette)
        if header is not None:
            self.header = header
        if row_header is not None:
            self.row_header = row_header
        if column_header is not None:
            self.column_header = column_header

    def __repr__(self):
        """Returns the string representation of the object."""
        return (
            f"MatrixChartInfo(name='{self.name}', type='{self.type}', header='{self.header}', row_header='"
            f"{self.row_header}', column_header='{self.column_header}')"
        )


_Metric = TypeVar("_Metric", bound=MetricEntity)
_VisualizationInfo = TypeVar("_VisualizationInfo", bound=VisualizationInfo)


class MetricsGroup(Generic[_Metric, _VisualizationInfo]):
    """This class aggregates a list of metric entities and defines how this group will be visualized on the UI.

    This class is the parent class to the different types of MetricsGroup that each represent a different type of chart
    in the UI.

    Example:
        An accuracy as a metrics group
            >>> acc = ScoreMetric("Accuracy", 0.5)
            >>> visual_info = BarChartInfo("Accuracy", visualization_type=_VisualizationInfo.BAR)  # show as radial bar
            >>> metrics_group = BarMetricsGroup([acc], visual_info)

        Loss curves as a metrics group
            >>> train_loss = CurveMetric("Train loss", xs=[0, 1, 2], ys=[5, 3, 1])
            >>> val_loss = CurveMetric("Validation", xs=[0, 1, 2], ys=[6, 4, 2])
            >>> visual_info = LineChartInfo("Loss curve", x_axis_label="# epoch", y_axis_label="Loss")
            >>> metrics_group = LineMetricsGroup([train_loss, val_loss], visual_info)
    """

    def __init__(self, metrics: Sequence[_Metric], visualization_info: _VisualizationInfo):
        if metrics is None or len(metrics) == 0:
            raise ValueError("Metrics cannot be None or empty")
        if visualization_info is None:
            raise ValueError("visualization_info cannot be None")
        self.metrics = metrics
        self.visualization_info = visualization_info


class MatrixMetricsGroup(MetricsGroup[MatrixMetric, MatrixChartInfo]):
    """This class represent a matrix chart in the UI.

    Multiple matrices can be displayed in the same chart.
    """

    def __init__(self, metrics: Sequence[MatrixMetric], visualization_info: MatrixChartInfo):
        super().__init__(metrics=metrics, visualization_info=visualization_info)


class LineMetricsGroup(MetricsGroup[CurveMetric, LineChartInfo]):
    """This class represent a line chart in the UI.

    Multiple lines can be displayed in a single chart.
    """

    def __init__(self, metrics: Sequence[CurveMetric], visualization_info: LineChartInfo):
        super().__init__(metrics=metrics, visualization_info=visualization_info)


class BarMetricsGroup(MetricsGroup[Union[ScoreMetric, CountMetric], BarChartInfo]):
    """This class represent a bar or radial bar chart in the UI.

    Each metric in the metrics group represents the value of a single bar/radial bar in the chart.
    """

    def __init__(
        self,
        metrics: Sequence[Union[ScoreMetric, CountMetric]],
        visualization_info: BarChartInfo,
    ):
        super().__init__(metrics=metrics, visualization_info=visualization_info)


class TextMetricsGroup(
    MetricsGroup[
        Union[ScoreMetric, CountMetric, InfoMetric, DateMetric, DurationMetric],
        TextChartInfo,
    ]
):
    """This class represent a text chart in the UI.

    Text charts contain only one metric, which can be of type CountMetric, ScoreMetric, DateMetric, DurationMetric or
    InfoMetric.
    """

    def __init__(
        self,
        metrics: Sequence[Union[ScoreMetric, CountMetric, InfoMetric, DateMetric, DurationMetric]],
        visualization_info: TextChartInfo,
    ):
        if not len(metrics) == 1:
            raise ValueError(
                "A text metrics group can contain only a single "
                "ScoreMetric, CountMetric, InfoMetric, DateMetric or "
                "DurationMetric."
            )
        super().__init__(metrics=metrics, visualization_info=visualization_info)


class Performance:
    """This performance class wraps the statistics of an entity (e.g., Model, Resultset).

    Args:
        score: the performance score. This will be the point of
            comparison between two performances.
        dashboard_metrics: (optional) additional statistics, containing
            charts, curves, and other additional info.
    """

    def __init__(self, score: ScoreMetric, dashboard_metrics: Optional[List[MetricsGroup]] = None):
        if not isinstance(score, ScoreMetric):
            raise ValueError(f"Expected score to be of type `ScoreMetric`, got type `{type(score)}` instead.")
        self._score: ScoreMetric = score
        self.dashboard_metrics: List[MetricsGroup] = [] if dashboard_metrics is None else dashboard_metrics

    @property
    def score(self):
        """Return the score metric."""
        return self._score

    def __eq__(self, other: object) -> bool:
        """Returns True if self and other have the same score and dashboard metrics."""
        if not isinstance(other, Performance):
            return False
        return self.score == other.score

    def __repr__(self):
        """Returns a string representation of the performance."""
        return f"Performance(score: {self.score.value}, dashboard: ({len(self.dashboard_metrics)} metric groups))"


class NullPerformance(Performance):
    """This is used to represent 'Performance not found'."""

    def __init__(self) -> None:
        super().__init__(score=ScoreMetric(name="Null score", value=0.0))

    def __repr__(self):
        """Returns a string representation of the performance."""
        return "NullPerformance()"

    def __eq__(self, other):
        """Returns True if other is a NullPerformance."""
        return isinstance(other, NullPerformance)


class MultiScorePerformance(Performance):
    """This class can be used in tasks where performance is measured by multiple metrics.

    Args:
        primary_score: The main performance score.
        additional_metrics: List of additional scores. When no primary
            score is provided, the first additional score takes priority
            as the main project score.
        dashboard_metrics: (optional) additional statistics, containing
            charts, curves, and other additional info.
    """

    def __init__(
        self,
        primary_score: Optional[ScoreMetric] = None,
        additional_scores: Optional[List[ScoreMetric]] = None,
        dashboard_metrics: Optional[List[MetricsGroup]] = None,
    ):
        assert primary_score is not None or (
            additional_scores is not None and len(additional_scores) > 0
        ), "Provide at least one primary or additional score."

        self._primary_score = primary_score
        self._additional_scores: List[ScoreMetric] = [] if additional_scores is None else additional_scores
        self.dashboard_metrics: List[MetricsGroup] = [] if dashboard_metrics is None else dashboard_metrics

        if self.primary_score is None:
            super().__init__(self.additional_scores[0], dashboard_metrics)
        else:
            super().__init__(self.primary_score, dashboard_metrics)

    @property
    def primary_score(self) -> Optional[ScoreMetric]:
        """Return the primary score metric."""
        return self._primary_score

    @property
    def additional_scores(self) -> List[ScoreMetric]:
        """Return the additional score metrics."""
        return self._additional_scores

    def __eq__(self, other: object) -> bool:
        """Returns True if the other object is a MultiScorePerformance with the same primary and additional scores."""
        if not isinstance(other, MultiScorePerformance):
            return False
        return self.primary_score == other.primary_score and self.additional_scores == other.additional_scores

    def __repr__(self):
        """Returns the representation of the performance."""
        return (
            f"MultiScorePerformance(score: {self.score.value}, primary_metric: {self.primary_score}, "
            f"additional_metrics: ({len(self.additional_scores)} metrics), "
            f"dashboard: ({len(self.dashboard_metrics)} metric groups))"
        )



"""Utils for demo."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import json
from pathlib import Path
from typing import Optional

from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model_template import TaskType, task_type_to_label_domain
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    create_converter,
)
from otx.api.usecases.exportable_code.visualizers import Visualizer


def get_model_path(path: Optional[Path]) -> Path:
    """Get path to model."""
    model_path = path
    if model_path is None:
        model_path = Path(__file__).parent / "model.xml"
    if not model_path.exists():
        raise IOError("The path to the model was not found.")

    return model_path


def get_parameters(path: Optional[Path]) -> dict:
    """Get hyper parameters to creating model."""
    parameters_path = path
    if parameters_path is None:
        parameters_path = Path(__file__).parent / "config.json"
    if not parameters_path.exists():
        raise IOError("The path to the config was not found.")

    with open(parameters_path, "r", encoding="utf8") as file:
        parameters = json.load(file)

    return parameters


def create_output_converter(task_type: TaskType, labels: LabelSchemaEntity):
    """Create annotation converter according to kind of task."""

    converter_type = task_type_to_label_domain(task_type)
    return create_converter(converter_type, labels)


def create_visualizer(_task_type: TaskType, no_show: bool = False):
    """Create visualizer according to kind of task."""

    # TODO: use anomaly-specific visualizer for anomaly tasks

    return Visualizer(window_name="Result", no_show=no_show)



"""NNCF config utils."""
# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
from copy import copy


def load_nncf_config(path):
    """loading_nncf_config."""
    assert path.endswith(".json"), f"Only json files are allowed as optimisation configs, provided {path}"
    with open(path, encoding="utf-8") as f_src:
        nncf_config = json.load(f_src)
    return nncf_config


def compose_nncf_config(nncf_config, enabled_options):
    """compose_nncf_config."""
    optimisation_parts = nncf_config

    if "order_of_parts" in optimisation_parts:
        # The result of applying the changes from optimisation parts
        # may depend on the order of applying the changes
        # (e.g. if for nncf_quantization it is sufficient to have `total_epochs=2`,
        #  but for sparsity it is required `total_epochs=50`)
        # So, user can define `order_of_parts` in the optimisation_config
        # to specify the order of applying the parts.
        order_of_parts = optimisation_parts["order_of_parts"]
        assert isinstance(order_of_parts, list), 'The field "order_of_parts" in optimisation config should be a list'

        for part in enabled_options:
            assert (
                part in order_of_parts
            ), f"The part {part} is selected, but it is absent in order_of_parts={order_of_parts}"

        optimisation_parts_to_choose = [part for part in order_of_parts if part in enabled_options]

    assert "base" in optimisation_parts, 'Error: the optimisation config does not contain the "base" part'
    nncf_config_part = optimisation_parts["base"]

    for part in optimisation_parts_to_choose:
        assert part in optimisation_parts, f'Error: the optimisation config does not contain the part "{part}"'
        optimisation_part_dict = optimisation_parts[part]
        try:
            nncf_config_part = merge_dicts_and_lists_b_into_a(nncf_config_part, optimisation_part_dict)
        except AssertionError as cur_error:
            err_descr = (
                f"Error during merging the parts of nncf configs:\n"
                f"the current part={part}, "
                f"the order of merging parts into base is {optimisation_parts_to_choose}.\n"
                f"The error is:\n{cur_error}"
            )
            raise RuntimeError(err_descr) from None

    return nncf_config_part


def merge_dicts_and_lists_b_into_a(a, b):  # pylint: disable=invalid-name
    """merge_dict."""
    return _merge_dicts_and_lists_b_into_a(a, b, "")


def _merge_dicts_and_lists_b_into_a(a, b, cur_key=None):  # pylint: disable=invalid-name
    """merge_dict inner function.

    The function is inspired by mmcf.Config._merge_a_into_b,
    but it
    * works with usual dicts and lists and derived types
    * supports merging of lists (by concatenating the lists)
    * makes recursive merging for dict + dict case
    * overwrites when merging scalar into scalar
    Note that we merge b into a (whereas Config makes merge a into b),
    since otherwise the order of list merging is counter-intuitive.
    """

    def _err_str(_a, _b, _key):
        if _key is None:
            _key_str = "of whole structures"
        else:
            _key_str = f"during merging for key=`{_key}`"
        return (
            f"Error in merging parts of config: different types {_key_str},"
            f" type(a) = {type(_a)},"
            f" type(b) = {type(_b)}"
        )

    assert isinstance(a, (dict, list)), f"Can merge only dicts and lists, whereas type(a)={type(a)}"
    assert isinstance(b, (dict, list)), _err_str(a, b, cur_key)
    assert isinstance(a, list) == isinstance(b, list), _err_str(a, b, cur_key)
    if isinstance(a, list):
        # the main diff w.r.t. mmcf.Config -- merging of lists
        return a + b

    a = copy(a)
    for k in b.keys():
        if k not in a:
            a[k] = copy(b[k])
            continue
        new_cur_key = cur_key + "." + k if cur_key else k
        if isinstance(a[k], (dict, list)):
            a[k] = _merge_dicts_and_lists_b_into_a(a[k], b[k], new_cur_key)
            continue

        assert not isinstance(b[k], (dict, list)), _err_str(a[k], b[k], new_cur_key)

        # suppose here that a[k] and b[k] are scalars, just overwrite
        a[k] = b[k]
    return a



"""This module implements helpers for converting shape entities."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.shapes.shape import ShapeEntity


class ShapeFactory:
    """Helper class converting between shape types."""

    @staticmethod
    def shape_as_rectangle(shape: ShapeEntity) -> Rectangle:
        """Get the outer-fitted rectangle representation of the shape.

        `media_width` and `media_height` are the width and height of the media in
        which the shape is expressed.

        Example:

            Let's assume a DatasetItem `dataset_item`.

            To obtain the shapes inside the full annotation as rectangles, one could call:

            >>> from otx.api.entities.dataset_item import DatasetItem
            >>> from otx.api.entities.image import NullImage
            >>> from otx.api.entities.annotation_scene import NullAnnotationScene
            >>> dataset_item = DatasetItem(media=NullImage(),
            >>>     annotation=NullAnnotationScene())
            >>> rectangles = [ShapeFactory.shape_as_rectangle(shape,
            >>>     dataset_item.media.width, dataset_item.media.height) for shape
            ... in dataset_item.annotation_scene.shapes]

            To obtain the shapes inside the dataset item (note that dataset item can have
            roi), one should call:

            >>> rectangles = [ShapeFactory.shape_as_rectangle(shape) for shape in
            >>>     dataset_item.get_annotations()]

            Since the shapes in the first call come from annotation directly, this means
            they are expressed in the media coordinate system. Therefore,
            dataset_item.media.width and dataset_item.media.height are passed.
            While in the second call, the shapes come from denormalization results wrt.
            dataset_item.roi and therefore expressed inside the roi. In this case,
            dataset_item.width and dataset_item.height are passed.

            Converting Ellipse to rectangle

            >>> height = 240
            >>> width = 480
            >>> rectangle = Rectangle(x1=0.375, y1=0.25, x2=0.625, y2=0.75,
            ... labels=[]) # a square of 120 x 120 pixels at the center of the image
            >>> ellipse = Ellipse(x1=0.5, y1=0.5, x2=0.625, y2=0.56125,
            ... labels=[]) # an ellipse of radius 60 pixels (x2 is wrt width) at the
            center of the image
            >>> ellipse_as_rectangle = ShapeFactory.shape_as_rectangle(ellipse)  # get the
            fitted rectangle for the ellipse
            >>> str(rectangle) == str(ellipse_as_rectangle)
            True

            Converting triangle to rectangle

            >>> points = [Point(x=0.5, y=0.25), Point(x=0.375, y=0.75),
            >>>     Point(x=0.625, y=0.75)]
            >>> triangle = Polygon(points=points, labels=[])
            >>> triangle_as_rectangle = ShapeFactory.shape_as_rectangle(triangle)
            >>> str(triangle_as_rectangle) == str(rectangle)
            True

        Args:
            shape (ShapeEntity): the shape to convert to rectangle

        Returns:
            Rectangle: bounding box of the shape
        """
        if isinstance(shape, Rectangle):
            return shape

        if isinstance(shape, Ellipse):
            x1 = shape.x1
            y1 = shape.y1
            x2 = shape.x2
            y2 = shape.y2
        elif isinstance(shape, Polygon):
            x1 = shape.min_x
            x2 = shape.max_x
            y1 = shape.min_y
            y2 = shape.max_y
        else:
            raise NotImplementedError(f"Conversion of a {type(shape)} to a rectangle is not implemented yet: {shape}")

        new_shape = Rectangle(x1=x1, y1=y1, x2=x2, y2=y2)
        return new_shape

    @staticmethod
    def shape_as_polygon(shape: ShapeEntity) -> Polygon:
        """Return a shape converted as polygon.

        For a rectangle, a polygon will be constructed with a point in each corner.
        For a ellipse, 360 points will be made. Otherwise, the original shape will be returned.
        The width/height for the parent need to be specified to make sure the aspect ratio is maintained.

        Args:
            shape (ShapeEntity): the shape to convert to polygon.

        Returns:
            Polygon: the polygon representation of the shape.

        Raises:
            NotImplementedError: if the shape is not a rectangle or ellipse.
        """
        if isinstance(shape, Polygon):
            new_shape = shape
        elif isinstance(shape, Rectangle):
            points = [
                Point(x=shape.x1, y=shape.y1),
                Point(x=shape.x2, y=shape.y1),
                Point(x=shape.x2, y=shape.y2),
                Point(x=shape.x1, y=shape.y2),
                Point(x=shape.x1, y=shape.y1),
            ]
            new_shape = Polygon(points=points)
        elif isinstance(shape, Ellipse):
            coordinates = shape.get_evenly_distributed_ellipse_coordinates()
            points = [Point(x=point[0], y=point[1]) for point in coordinates]
            new_shape = Polygon(points=points)
        else:
            raise NotImplementedError(f"Conversion of a {type(shape)} to a polygon is not implemented yet: " f"{shape}")
        return new_shape

    @staticmethod
    def shape_as_ellipse(shape: ShapeEntity) -> Ellipse:
        """Returns the inner-fitted ellipse for a given shape.

        Args:
            shape (ShapeEntity): Shape to convert.

        Returns:
            Ellipse: Ellipse representation of the shape.

        Raises:
            NotImplementedError: If the shape is not a rectangle or polygon.
        """
        if isinstance(shape, Ellipse):
            return shape

        if isinstance(shape, Rectangle):
            x1 = shape.x1
            x2 = shape.x2
            y1 = shape.y1
            y2 = shape.y2
        elif isinstance(shape, Polygon):
            x1 = shape.min_x
            x2 = shape.max_x
            y1 = shape.min_y
            y2 = shape.max_y
        else:
            raise NotImplementedError(f"Conversion of a {type(shape)} to an ellipse is not implemented yet: {shape}")
        return Ellipse(x1=x1, y1=y1, x2=x2, y2=y2)

    @staticmethod
    def shape_produces_valid_crop(shape: ShapeEntity, media_width: int, media_height: int) -> bool:
        """Check if crop is valid.

        Checks if the shape produces a valid crop based on the image width and height,
        regardless of the contents of the image.

        Args:
            shape (ShapeEntity): Shape to check
            media_width (int): Width of the image
            media_height (int): Height of the image

        Returns:
            bool: True if the shape produces a valid crop, False otherwise
        """
        is_valid = True
        try:
            shape_as_rectangle = ShapeFactory.shape_as_rectangle(shape=shape)
        except ValueError:
            # Thrown if the resulting bounding box is invalid
            return False

        # The min and max are to handle out-of-bounds coordinates
        x1 = min(max(round(shape_as_rectangle.x1 * media_width), 0), media_width)
        x2 = min(max(round(shape_as_rectangle.x2 * media_width), 0), media_width)
        y1 = min(max(round(shape_as_rectangle.y1 * media_height), 0), media_height)
        y2 = min(max(round(shape_as_rectangle.y2 * media_height), 0), media_height)

        if x1 == x2 or y1 == y2:
            is_valid = False
        return is_valid



"""This module define the color entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import abc
import random
from typing import Tuple


class ColorEntity(metaclass=abc.ABCMeta):
    """This class represents an abstract Color, some functions are still abstract."""

    def __init__(self, red: int, green: int, blue: int, alpha: int):
        self.__red = red
        self.__green = green
        self.__blue = blue
        self.__alpha = alpha

    @property
    def red(self) -> int:
        """Returns the red color value for the ColorEntity object."""
        return self.__red

    @red.setter
    def red(self, value):
        self.__red = value

    @property
    def green(self) -> int:
        """Returns the green color value for the ColorEntity object."""
        return self.__green

    @green.setter
    def green(self, value):
        self.__green = value

    @property
    def blue(self) -> int:
        """Returns the blue color value for the ColorEntity object."""
        return self.__blue

    @blue.setter
    def blue(self, value):
        self.__blue = value

    @property
    def alpha(self) -> int:
        """Returns the alpha value for the ColorEntity object."""
        return self.__alpha

    @alpha.setter
    def alpha(self, value):
        self.__alpha = value

    @property
    @abc.abstractmethod
    def hex_str(self) -> str:
        """Returns the color in a Hex representation."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_hex_str(cls, string: str):
        """Converts a hex string to a color.

        Args:
            string (str): The hex string
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def random(cls):
        """Generates a random Color."""
        raise NotImplementedError


class Color(ColorEntity):
    """Represents an RGBA color."""

    def __init__(self, red: int, green: int, blue: int, alpha: int = 255):
        super().__init__(red=red, green=green, blue=blue, alpha=alpha)

    def __repr__(self):
        """Returns string representation of the color."""
        return f"Color(red={self.red}, green={self.green}, blue={self.blue}, alpha={self.alpha})"

    def __eq__(self, other):
        """Returns True if both colors are equal."""
        if isinstance(other, Color):
            return (
                self.red == other.red
                and self.green == other.green
                and self.blue == other.blue
                and self.alpha == other.alpha
            )
        return False

    @property
    def hex_str(self) -> str:
        """Returns the color in a Hex representation."""
        return f"#{self.red:02x}{self.green:02x}{self.blue:02x}{self.alpha:02x}"

    @classmethod
    def from_hex_str(cls, string: str) -> "Color":
        """Creates Color() instance given a hex string.

        Supports 6 character hex string (RGB), or 8 character hex string (RGBA).
        The string might optionally start with a number sign (#).

        Example:
            Creating color object:

            >>> Color.from_hex_str("#ff0000")
            Color(red=255, green=0, blue=0, alpha=255)

            >>> Color.from_hex_str("0000ff")
            Color(red=0, green=0, blue=255, alpha=255)

            >>> Color.from_hex_str("#96Ff00C8")
            Color(red=150, green=255, blue=0, alpha=200)

        Args:
            string (str): Hex string

        Returns:
            Color instance
        """
        string = string.lstrip("#").lower()
        if len(string) < 8:
            # If alpha channel misses, add it
            string += (8 - len(string)) * "f"
        red, green, blue, alpha = tuple(int(string[i : i + 2], 16) for i in (0, 2, 4, 6))
        return cls(red=red, green=green, blue=blue, alpha=alpha)

    @classmethod
    def random(cls) -> "Color":
        """Generate random Color() instance.

        Returns:
            Color instance with random color
        """
        # This random is not used for security/cryptographic purposes, so we can ignore this possible security risk
        red, green, blue = (
            random.randint(0, 255),  # nosec
            random.randint(0, 255),  # nosec
            random.randint(0, 255),  # nosec
        )
        return cls(red=red, green=green, blue=blue, alpha=255)

    @property
    def rgb_tuple(self) -> Tuple[int, int, int]:
        """Retrieves the Color as a RGB tuple.

        Returns:
            Tuple[int, int, int]
        """
        return self.red, self.green, self.blue

    @property
    def bgr_tuple(self) -> Tuple[int, int, int]:
        """Retrieves the Color as a BGR tuple.

        Returns:
            Tuple[int, int, int]
        """
        return self.blue, self.green, self.red



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch
from mmdet.core import build_anchor_generator
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.ssd_head import SSDHead

from ...mmov_model import MMOVModel


@HEADS.register_module()
class MMOVSSDHead(SSDHead):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
        transpose_cls: bool = False,
        transpose_reg: bool = False,
        background_index: Optional[int] = None,
        *args,
        **kwargs,
    ):

        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._inputs = deepcopy(inputs)
        self._outputs = deepcopy(outputs)
        self._init_weight = init_weight
        self._verify_shape = verify_shape
        self._transpose_cls = transpose_cls
        self._transpose_reg = transpose_reg
        self._background_index = background_index

        # dummy input
        anchor_generator = build_anchor_generator(kwargs["anchor_generator"])
        num_anchors = anchor_generator.num_base_anchors
        in_channels = [256 for _ in num_anchors]
        super().__init__(in_channels=in_channels, *args, **kwargs)

        self.cls_convs = torch.nn.ModuleList()
        self.reg_convs = torch.nn.ModuleList()

        for (
            inputs,
            outputs,
        ) in zip(self._inputs["cls_convs"], self._outputs["cls_convs"]):
            self.cls_convs.append(
                MMOVModel(
                    self._model_path_or_model,
                    self._weight_path,
                    inputs=inputs,
                    outputs=outputs,
                    remove_normalize=False,
                    merge_bn=False,
                    paired_bn=False,
                    init_weight=self._init_weight,
                    verify_shape=self._verify_shape,
                )
            )

        for (
            inputs,
            outputs,
        ) in zip(self._inputs["reg_convs"], self._outputs["reg_convs"]):
            self.reg_convs.append(
                MMOVModel(
                    self._model_path_or_model,
                    self._weight_path,
                    inputs=inputs,
                    outputs=outputs,
                    remove_normalize=False,
                    merge_bn=False,
                    paired_bn=False,
                    init_weight=self._init_weight,
                    verify_shape=self._verify_shape,
                )
            )

    def forward(self, feats):
        cls_scores = []
        bbox_preds = []
        for feat, reg_conv, cls_conv in zip(feats, self.reg_convs, self.cls_convs):

            cls_score = cls_conv(feat)
            bbox_pred = reg_conv(feat)

            if self._transpose_cls:
                # [B, cls_out_channels * num_anchors, H, W]
                #   -> [B, num_anchors * cls_out_channels, H, W]
                shape = cls_score.shape
                cls_score = (
                    cls_score.reshape(shape[0], self.cls_out_channels, -1, *shape[2:]).transpose(1, 2).reshape(shape)
                )

            if self._transpose_reg:
                # [B, 4 * num_anchors, H, W] -> [B, num_anchors * 4, H, W]
                shape = bbox_pred.shape
                bbox_pred = bbox_pred.reshape(shape[0], 4, -1, *shape[2:]).transpose(1, 2).reshape(shape)

            # since mmdet v2.0, SSDHead is supposed to be
            # that FG labels to [0, num_class-1] and BG labels to num_class
            # but ssd300, ssd512, etc. from OMZ are
            # that FG labels to [1, num_class] and BG labels to 0
            if self._background_index is not None and cls_score is not None:
                cls_score = cls_score.permute(0, 2, 3, 1)
                shape = cls_score.shape
                cls_score = cls_score.reshape(-1, self.cls_out_channels)
                cls_score = torch.cat(
                    (
                        cls_score[:, : self._background_index],
                        cls_score[:, self._background_index + 1 :],
                        cls_score[:, self._background_index : self._background_index + 1],
                    ),
                    -1,
                )
                cls_score = cls_score.reshape(shape)
                cls_score = cls_score.permute(0, 3, 1, 2)

            cls_scores.append(cls_score)
            bbox_preds.append(bbox_pred)
        return cls_scores, bbox_preds

    def init_weights(self):
        # TODO
        pass



"""OTX Adapters - mmdet.utils."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .builder import build_detector
from .config_utils import (
    cluster_anchors,
    patch_config,
    patch_datasets,
    patch_evaluation,
    prepare_for_training,
    set_hyperparams,
)

__all__ = [
    "cluster_anchors",
    "patch_config",
    "patch_datasets",
    "patch_evaluation",
    "prepare_for_training",
    "set_hyperparams",
    "build_detector",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv import ConfigDict

from otx.mpa.det.stage import DetectionStage
from otx.mpa.utils.config_utils import update_or_add_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class SemiSLDetectionStage(DetectionStage):
    """Patch config to support semi supervised learning for object detection"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def configure_data(self, cfg, training, data_cfg, **kwargs):
        """Patch cfg.data."""
        super().configure_data(cfg, training, data_cfg, **kwargs)
        # Set unlabeled data hook
        if training:
            if cfg.data.get("unlabeled", False) and cfg.data.unlabeled.get("otx_dataset", False):
                if len(cfg.data.unlabeled.get("pipeline", [])) == 0:
                    cfg.data.unlabeled.pipeline = cfg.data.train.pipeline.copy()
                self.configure_unlabeled_dataloader(cfg, self.distributed)

    def configure_task(self, cfg, training, **kwargs):
        """Patch config to support training algorithm."""
        logger.info(f"Semi-SL task config!!!!: training={training}")
        if "task_adapt" in cfg:
            self.task_adapt_type = cfg["task_adapt"].get("type", None)
            self.task_adapt_op = cfg["task_adapt"].get("op", "REPLACE")
            self.configure_classes(cfg)

            if self.data_classes != self.model_classes:
                self.configure_task_data_pipeline(cfg)
            # TODO[JAEGUK]: configure_anchor is not working
            if cfg["task_adapt"].get("use_mpa_anchor", False):
                self.configure_anchor(cfg)
            if self.task_adapt_type == "mpa":
                self.configure_bbox_head(cfg)
            else:
                src_data_cfg = self.get_data_cfg(cfg, "train")
                src_data_cfg.pop("old_new_indices", None)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import otx.mpa.modules.datasets.pipelines.torchvision2mmdet
import otx.mpa.modules.datasets.task_adapt_dataset
import otx.mpa.modules.hooks
import otx.mpa.modules.hooks.composed_dataloaders_hook
import otx.mpa.modules.models.backbones.imgclsmob
import otx.mpa.modules.models.detectors
import otx.mpa.modules.models.heads.cross_dataset_detector_head
import otx.mpa.modules.models.heads.custom_anchor_generator
import otx.mpa.modules.models.heads.custom_atss_head
import otx.mpa.modules.models.heads.custom_retina_head
import otx.mpa.modules.models.heads.custom_roi_head
import otx.mpa.modules.models.heads.custom_ssd_head
import otx.mpa.modules.models.heads.custom_vfnet_head
import otx.mpa.modules.models.heads.custom_yolox_head
import otx.mpa.modules.models.losses.cross_focal_loss
import otx.mpa.modules.models.losses.l2sp_loss

# flake8: noqa
from . import explainer, exporter, incremental, inferrer, semisl, stage, trainer



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import functools
from collections import OrderedDict

import numpy as np
import torch
from mmdet.models import DETECTORS, build_detector
from mmdet.models.detectors import BaseDetector

from otx.mpa.utils.logger import get_logger

from .sam_detector_mixin import SAMDetectorMixin

logger = get_logger()


@DETECTORS.register_module()
class UnbiasedTeacher(SAMDetectorMixin, BaseDetector):
    """Unbiased teacher frameowork for general detectors"""

    def __init__(
        self,
        unlabeled_loss_weight=1.0,
        unlabeled_loss_names=[
            "loss_cls",
        ],
        pseudo_conf_thresh=0.7,
        enable_unlabeled_loss=False,
        bg_loss_weight=-1.0,
        **kwargs
    ):
        super().__init__()
        self.unlabeled_loss_weight = unlabeled_loss_weight
        self.unlabeled_loss_names = unlabeled_loss_names
        self.pseudo_conf_thresh = pseudo_conf_thresh
        self.unlabeled_loss_enabled = enable_unlabeled_loss
        self.bg_loss_weight = bg_loss_weight

        cfg = kwargs.copy()
        arch_type = cfg.pop("arch_type")
        cfg["type"] = arch_type
        self.model_s = build_detector(cfg)
        self.model_t = copy.deepcopy(self.model_s)

        # Hooks for super_type transparent weight load/save
        self._register_state_dict_hook(self.state_dict_hook)
        self._register_load_state_dict_pre_hook(functools.partial(self.load_state_dict_pre_hook, self))

    def extract_feat(self, imgs):
        return self.model_t.extract_feat(imgs)

    def simple_test(self, img, img_metas, **kwargs):
        return self.model_t.simple_test(img, img_metas, **kwargs)

    def aug_test(self, imgs, img_metas, **kwargs):
        return self.model_t.aug_test(imgs, img_metas, **kwargs)

    def forward_dummy(self, img, **kwargs):
        return self.model_t.forward_dummy(img, **kwargs)

    def enable_unlabeled_loss(self, mode=True):
        self.unlabeled_loss_enabled = mode

    def forward_train(self, img, img_metas, img0, gt_bboxes, gt_labels, gt_bboxes_ignore=None, **kwargs):

        losses = {}

        # Supervised loss
        # TODO: check img0 only option (which is common for mean teacher method)
        sl_losses = self.model_s.forward_train(
            torch.cat((img0, img)),  # weak + hard augmented images
            img_metas + img_metas,
            gt_bboxes + gt_bboxes,
            gt_labels + gt_labels,
            gt_bboxes_ignore + gt_bboxes_ignore if gt_bboxes_ignore else None,
        )
        losses.update(sl_losses)

        # Pseudo labels from teacher
        ul_args = kwargs.get("extra_0", {})  # Supposing ComposedDL([labeled, unlabeled]) data loader
        ul_img = ul_args.get("img")
        ul_img0 = ul_args.get("img0")
        ul_img_metas = ul_args.get("img_metas")
        if ul_img is None:
            return losses
        with torch.no_grad():
            teacher_outputs = self.model_t.forward_test(
                [ul_img0],
                [ul_img_metas],
                rescale=False,  # easy augmentation
            )
        pseudo_bboxes, pseudo_labels, pseudo_ratio = self.generate_pseudo_labels(
            teacher_outputs, device=ul_img0.device, **kwargs
        )
        ps_recall = self.eval_pseudo_label_recall(pseudo_bboxes, ul_args.get("gt_bboxes", []))
        losses.update(ps_recall=ps_recall)
        losses.update(ps_ratio=torch.Tensor([pseudo_ratio]))

        if not self.unlabeled_loss_enabled or self.unlabeled_loss_weight <= 0.001:  # TODO: move back
            return losses

        # Unsupervised loss
        if self.bg_loss_weight >= 0.0:
            self.model_s.bbox_head.bg_loss_weight = self.bg_loss_weight
        ul_losses = self.model_s.forward_train(ul_img, ul_img_metas, pseudo_bboxes, pseudo_labels)  # hard augmentation
        if self.bg_loss_weight >= 0.0:
            self.model_s.bbox_head.bg_loss_weight = -1.0

        for ul_loss_name in self.unlabeled_loss_names:
            ul_loss = ul_losses[ul_loss_name]
            if isinstance(ul_loss, torch.Tensor):
                ul_loss = [ul_loss]
            losses[ul_loss_name + "_ul"] = [loss * self.unlabeled_loss_weight for loss in ul_loss]
        # TODO: apply loss_bbox when adopting QFL;

        return losses

    def generate_pseudo_labels(self, teacher_outputs, **kwargs):
        device = kwargs.pop("device")
        all_pseudo_bboxes = []
        all_pseudo_labels = []
        num_all_bboxes = 0
        num_all_pseudo = 0
        for teacher_bboxes_labels in teacher_outputs:
            pseudo_bboxes = []
            pseudo_labels = []
            for label, teacher_bboxes in enumerate(teacher_bboxes_labels):
                confidences = teacher_bboxes[:, -1]
                pseudo_indices = confidences > self.pseudo_conf_thresh
                pseudo_bboxes.append(teacher_bboxes[pseudo_indices, :4])  # model output: [x y w h conf]
                pseudo_labels.append(np.full([sum(pseudo_indices)], label))
                num_all_bboxes += teacher_bboxes.shape[0]
                num_all_pseudo += pseudo_bboxes[-1].shape[0]
            all_pseudo_bboxes.append(torch.from_numpy(np.concatenate(pseudo_bboxes)).to(device))
            all_pseudo_labels.append(torch.from_numpy(np.concatenate(pseudo_labels)).to(device))
        # print(f'{num_all_pseudo} / {num_all_bboxes}')
        pseudo_ratio = float(num_all_pseudo) / num_all_bboxes if num_all_bboxes > 0 else 0.0
        return all_pseudo_bboxes, all_pseudo_labels, pseudo_ratio

    def eval_pseudo_label_recall(self, all_pseudo_bboxes, all_gt_bboxes):
        # For test only
        from mmdet.core.evaluation.recall import _recalls, bbox_overlaps

        img_num = len(all_gt_bboxes)
        if img_num == 0:
            return torch.Tensor([0.0])
        all_ious = np.ndarray((img_num,), dtype=object)
        for i in range(img_num):
            ps_bboxes = all_pseudo_bboxes[i]
            gt_bboxes = all_gt_bboxes[i]
            # prop_num = min(ps_bboxes.shape[0], 100)
            prop_num = ps_bboxes.shape[0]
            if gt_bboxes is None or gt_bboxes.shape[0] == 0:
                ious = np.zeros((0, ps_bboxes.shape[0]), dtype=np.float32)
            elif ps_bboxes is None or ps_bboxes.shape[0] == 0:
                ious = np.zeros((gt_bboxes.shape[0], 0), dtype=np.float32)
            else:
                ious = bbox_overlaps(gt_bboxes.detach().cpu().numpy(), ps_bboxes.detach().cpu().numpy()[:prop_num, :4])
            all_ious[i] = ious
        recall = _recalls(all_ious, np.array([100]), np.array([0.5]))
        return torch.Tensor(recall)

    @staticmethod
    def state_dict_hook(module, state_dict, prefix, *args, **kwargs):
        """Redirect teacher model as output state_dict (student as auxilliary)"""
        logger.info("----------------- UnbiasedTeacher.state_dict_hook() called")
        for k in list(state_dict.keys()):
            v = state_dict.pop(k)
            if not prefix or k.startswith(prefix):
                k = k.replace(prefix, "", 1)
                if k.startswith("model_t."):
                    k = k.replace("model_t.", "", 1)
                elif k.startswith("model_s."):
                    continue
                k = prefix + k
            state_dict[k] = v
        return state_dict

    @staticmethod
    def load_state_dict_pre_hook(module, state_dict, *args, **kwargs):
        """Redirect input state_dict to teacher model"""
        logger.info("----------------- UnbiasedTeacher.load_state_dict_pre_hook() called")
        for k in list(state_dict.keys()):
            v = state_dict.pop(k)
            if "model_s." not in k:
                k = "model_t." + k
            state_dict[k] = v



"""Detection models static deploy config."""

ir_config = dict(
    type="onnx",
    export_params=True,
    keep_initializers_as_inputs=False,
    opset_version=11,
    save_file="end2end.onnx",
    input_names=["image"],
    output_names=["boxes", "labels"],
    input_shape=None,
    # TODO
    # optimizing onnx graph mess up NNCF graph at some point
    # where we need to look into
    optimize=False,
)

codebase_config = dict(
    type="mmdet",
    task="ObjectDetection",
    model_type="end2end",
    post_processing=dict(
        score_threshold=0.05,
        confidence_threshold=0.005,  # for YOLOv3
        iou_threshold=0.5,
        max_output_boxes_per_class=200,
        pre_top_k=5000,
        keep_top_k=100,
        background_label_id=-1,
    ),
)

backend_config = dict(
    type="openvino",
    mo_options=None,
)

input_data = dict(
    shape=(128, 128, 3),
    file_path=None,
)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from abc import ABC, abstractmethod
from collections import OrderedDict
from pprint import pformat
from typing import Any, Callable, Dict, Optional

import pytest

from .e2e_test_system import DataCollector
from .logging import get_logger
from .training_tests_actions import BaseOTXTestAction
from .training_tests_common import REALLIFE_USECASE_CONSTANT

logger = get_logger()


def get_value_from_dict_by_dot_separated_address(struct, address):
    def _get(cur_struct, addr):
        assert isinstance(addr, list)
        if not addr:
            return cur_struct
        assert isinstance(cur_struct, dict)
        if addr[0] not in cur_struct:
            raise ValueError(f"Cannot find address {address} in struct {struct}: {addr[0]} is absent in {cur_struct}")
        return _get(cur_struct[addr[0]], addr[1:])

    assert isinstance(address, str), f"The parameter address should be string, address={address}"
    return _get(struct, address.split("."))


class Validator:
    """
    The class receives info on results metric of the current test stage and
    compares it with the expected metrics.
    """

    def __init__(self, cur_test_expected_metrics_callback: Optional[Callable[[], Dict]]):
        self.cur_test_expected_metrics_callback = cur_test_expected_metrics_callback

    # TODO(lbeynens): add a method to extract dependency info from expected metrics
    #                 to add the stages we depend on to the dependency list.

    @staticmethod
    def _get_min_max_value_from_expected_metrics(cur_metric_requirements: Dict, test_results_storage: Dict):
        """
        The method gets requirement for some metric and convert it to the triplet
        (target_value, min_value, max_value).
        Note that the target_value may be pointed either by key 'target_value' (in this case it is float),
        or by the key 'base', in this case it is a dot-separated address to another value in the
        storage of previous stages' results `test_results_storage`.

        Note that the range for the metric values may be pointed by key 'max_diff',
        in this case the range will be [target_value - max_diff, target_value + max_diff]
        (inclusively).

        But also the range may be pointed by keys 'max_diff_if_less_threshold' and
        'max_diff_if_greater_threshold', in this case the range is
        [target_value - max_diff_if_less_threshold, target_value + max_diff_if_greater_threshold]
        (also inclusively). This allows to point non-symmetric ranges w.r.t. the target_value.

        Also note that if one of 'max_diff_if_less_threshold' and 'max_diff_if_greater_threshold'
        is absent, it is set to `+infinity`, so the range will be bounded from one side
        (but not both of them, this will be an error)
        """
        keys = set(cur_metric_requirements.keys())
        if "target_value" not in keys and "base" not in keys:
            raise ValueError(
                f'Wrong cur_metric_requirements: either "target_value" or "base" '
                f" should be pointed in the structure, whereas "
                f"cur_metric_requirements={pformat(cur_metric_requirements)}"
            )
        if "target_value" in keys and "base" in keys:
            raise ValueError(
                f'Wrong cur_metric_requirements: either "target_value" or "base" '
                f" should be pointed in the structure, but not both, whereas "
                f"cur_metric_requirements={pformat(cur_metric_requirements)}"
            )
        if (
            ("max_diff" not in keys)
            and ("max_diff_if_less_threshold" not in keys)
            and ("max_diff_if_greater_threshold" not in keys)
        ):
            raise ValueError(
                f'Wrong cur_metric_requirements: either "max_diff" or one/two of '
                f'"max_diff_if_less_threshold" and "max_diff_if_greater_threshold" should be '
                f"pointed in the structure, whereas "
                f"cur_metric_requirements={pformat(cur_metric_requirements)}"
            )

        if ("max_diff" in keys) and ("max_diff_if_less_threshold" in keys or "max_diff_if_greater_threshold" in keys):
            raise ValueError(
                f'Wrong cur_metric_requirements: either "max_diff" or one/two of '
                f'"max_diff_if_less_threshold" and "max_diff_if_greater_threshold" should be '
                f"pointed in the structure, but not both, whereas "
                f"cur_metric_requirements={pformat(cur_metric_requirements)}"
            )

        if "target_value" in cur_metric_requirements:
            target_value = float(cur_metric_requirements["target_value"])
        elif "base" in cur_metric_requirements:
            base_metric_address = cur_metric_requirements["base"]
            target_value = get_value_from_dict_by_dot_separated_address(test_results_storage, base_metric_address)
            target_value = float(target_value)
        else:
            raise RuntimeError(f"ERROR: Wrong parsing of metric requirements {cur_metric_requirements}")

        if "max_diff" in cur_metric_requirements:
            max_diff = cur_metric_requirements["max_diff"]
            max_diff = float(max_diff)
            if not max_diff >= 0:
                raise ValueError(f"Wrong max_diff {max_diff} -- it should be a non-negative number")
            return (target_value, target_value - max_diff, target_value + max_diff)

        max_diff_if_less_threshold = cur_metric_requirements.get("max_diff_if_less_threshold")
        max_diff_if_greater_threshold = cur_metric_requirements.get("max_diff_if_greater_threshold")
        if max_diff_if_less_threshold is None and max_diff_if_greater_threshold is None:
            raise ValueError(
                f"Wrong cur_metric_requirements: all of max_diff, max_diff_if_less_threshold, and "
                f"max_diff_if_greater_threshold are None, "
                f"cur_metric_requirements={pformat(cur_metric_requirements)}"
            )

        if max_diff_if_greater_threshold is not None:
            max_diff_if_greater_threshold = float(max_diff_if_greater_threshold)
            if not max_diff_if_greater_threshold >= 0:
                raise ValueError(
                    f"Wrong max_diff_if_greater_threshold {max_diff_if_greater_threshold} "
                    f"-- it should be a non-negative number"
                )

            max_value = target_value + max_diff_if_greater_threshold
        else:
            max_value = None

        if max_diff_if_less_threshold is not None:
            max_diff_if_less_threshold = float(max_diff_if_less_threshold)
            if not max_diff_if_less_threshold >= 0:
                raise ValueError(
                    f"Wrong max_diff_if_less_threshold {max_diff_if_less_threshold} "
                    f"-- it should be a non-negative number"
                )

            min_value = target_value - max_diff_if_less_threshold
        else:
            min_value = None

        return (target_value, min_value, max_value)

    @staticmethod
    def _compare(
        current_metric: float,
        cur_res_addr: str,
        target_value: float,
        min_value: Optional[float],
        max_value: Optional[float],
    ):
        assert all(isinstance(v, float) for v in [current_metric, target_value])
        assert all((v is None) or isinstance(v, float) for v in [min_value, max_value])

        if min_value is not None and max_value is not None:
            assert min_value <= target_value <= max_value

            if min_value <= current_metric <= max_value:
                logger.info(
                    f"Validation: passed: The metric {cur_res_addr} is in the acceptable range "
                    f"near the target value {target_value}: "
                    f"{current_metric} is in [{min_value}, {max_value}]"
                )
                is_passed = True
                cur_fail_reason = None
            else:
                cur_fail_reason = (
                    f"Validation: failed: The metric {cur_res_addr} is NOT in the acceptable range "
                    f"near the target value {target_value}: "
                    f"{current_metric} is NOT in [{min_value}, {max_value}]"
                )
                logger.error(cur_fail_reason)
                is_passed = False
            return is_passed, cur_fail_reason

        assert (min_value is not None) or (max_value is not None)
        if min_value is not None:
            cmp_op = lambda x: x >= min_value  # noqa: E731
            cmp_str_true = "greater or equal"
            cmp_op_str_true = ">="
            cmp_op_str_false = "<"
            threshold = min_value
        else:
            assert max_value is not None
            cmp_op = lambda x: x <= max_value  # noqa: E731
            cmp_str_true = "less or equal"
            cmp_op_str_true = "<="
            cmp_op_str_false = ">"
            threshold = max_value
        acceptable_error = abs(threshold - target_value)
        if cmp_op(current_metric):
            logger.info(
                f"Validation: passed: The metric {cur_res_addr} is {cmp_str_true} "
                f"the target value {target_value} with acceptable error {acceptable_error}: "
                f"{current_metric} {cmp_op_str_true} {threshold}"
            )
            is_passed = True
            cur_fail_reason = None
        else:
            cur_fail_reason = (
                f"Validation: failed: The metric {cur_res_addr} is NOT {cmp_str_true} "
                f"the target value {target_value} with acceptable error {acceptable_error}: "
                f"{current_metric} {cmp_op_str_false} {threshold}"
            )
            logger.error(cur_fail_reason)
            is_passed = False
        return is_passed, cur_fail_reason

    def validate(self, current_result: Dict, test_results_storage: Dict):
        """
        The method validates results of the current test.
        :param current_result -- dict with result of the current test
        :param test_results_storage -- dict with results of previous tests
                                       of this test case
                                       (e.g. the same training parameters)

        The function returns nothing, but may raise exceptions to fail the test.
        If the structure stored expected metrics is wrong, the function raises ValueError.
        """
        if self.cur_test_expected_metrics_callback is None:
            # most probably, it is not a reallife test
            logger.info(
                f"Validation: skipped, since there should not be expected metrics for this test, "
                f'most probably the test is not run in "{REALLIFE_USECASE_CONSTANT}" usecase'
            )
            return

        logger.info("Validation: begin")

        # calling the callback to receive expected metrics for the current test
        cur_test_expected_metrics = self.cur_test_expected_metrics_callback()

        assert isinstance(
            cur_test_expected_metrics, dict
        ), f"Wrong current test expected metric: {cur_test_expected_metrics}"
        logger.debug(f"Validation: received cur_test_expected_metrics={pformat(cur_test_expected_metrics)}")
        is_passed = True
        fail_reasons = []
        for k, v in cur_test_expected_metrics.items():
            # TODO(lbeynens): add possibility to point a list of requirements for a metric
            cur_res_addr = k
            cur_metric_requirements = v
            logger.info(f"Validation: begin check {cur_res_addr}")
            try:
                current_metric = get_value_from_dict_by_dot_separated_address(current_result, cur_res_addr)
                current_metric = float(current_metric)
            except (ValueError, TypeError) as e:
                raise ValueError(f"Cannot get metric {cur_res_addr} from the current result {current_result}") from e

            logger.debug(f"current_metric = {current_metric}")
            try:
                (
                    target_value,
                    min_value,
                    max_value,
                ) = self._get_min_max_value_from_expected_metrics(cur_metric_requirements, test_results_storage)
            except (ValueError, TypeError) as e:
                raise ValueError(f"Error when parsing expected metrics for the metric {cur_res_addr}") from e

            cur_is_passed, cur_fail_reason = self._compare(
                current_metric, cur_res_addr, target_value, min_value, max_value
            )
            if not cur_is_passed:
                is_passed = False
                fail_reasons.append(cur_fail_reason)

            logger.info(f"Validation: end check {cur_res_addr}")

        logger.info(f"Validation: end, result={is_passed}")
        if not is_passed:
            fail_reasons_str = "\n".join(fail_reasons)
            pytest.fail(f"Validation failed:\n{fail_reasons_str}")


class OTXTestStagesStorageInterface(ABC):
    @abstractmethod
    def get_stage(self, name: str) -> "OTXTestStage":
        raise NotImplementedError("The method get_stage is not implemented")


class OTXTestStage:
    """
    OTXTestStage -- auxiliary class that
    1. Allows to set up dependency between test stages: before the main action of a test stage is run, all the actions
       for the stages that are pointed in 'depends' list are called beforehand;
    2. Runs for each test stage its main action only once: the main action is run inside try-except clause, and
       2.1. if the action was executed without exceptions, a flag `was_processed` is set, the results of the action
            are kept, and the next time the stage is called no action is executed;
       2.2. if the action raised an exception, the exception is stored, the flag `was_processed` is set, and the next
            time the stage is called the exception is re-raised.
    """

    def __init__(self, action: BaseOTXTestAction, stages_storage: OTXTestStagesStorageInterface):
        self.was_processed = False
        self.stored_exception: Optional[Exception] = None
        self.action = action
        self.stages_storage = stages_storage
        self.stage_results: Dict[str, Any] = {}
        assert isinstance(self.stages_storage, OTXTestStagesStorageInterface)
        assert isinstance(self.action, BaseOTXTestAction)

    def __str__(self):
        return (
            f"{type(self).__name__}("
            f"action={self.action}, "
            f"was_processed={self.was_processed}, "
            f"stored_exception={self.stored_exception}, "
            f"stage_results.keys={list(self.stage_results.keys())}, "
            f"id(stages_storage)={id(self.stages_storage)}"
            f")"
        )

    @property
    def name(self):
        return self.action.name

    def get_depends_stages(self):
        logger.debug(f"get_depends_stages for stage {self.name}: begin")
        depends_stages_names = self.action.depends_stages_names
        assert isinstance(depends_stages_names, list)
        assert all(isinstance(v, str) for v in depends_stages_names)

        stages = []
        for stage_name in depends_stages_names:
            logger.debug(f'get_depends_stages: get stage with name "{stage_name}"')
            cur_stage = self.stages_storage.get_stage(stage_name)
            assert isinstance(cur_stage, OTXTestStage), f'Wrong stage for stage_name="{stage_name}"'
            assert (
                cur_stage.name == stage_name
            ), f'For stage_name="{stage_name}" got the stage with name="{cur_stage.name}"'
            logger.debug(f'get_depends_stages: cur_stage="{cur_stage}"')
            stages.append(cur_stage)
        logger.debug(f"get_depends_stages for stage {self.name}: end")
        return stages

    def _reraise_stage_exception_if_was_failed(self):
        assert (
            self.was_processed
        ), "The method _reraise_stage_exception_if_was_failed should be used only for stages that were processed"
        if self.stored_exception is None:
            # nothing to do here
            return

        logger.warning(
            f"In stage {self.name}: found that previous call of the stage " "caused exception -- re-raising it"
        )
        raise self.stored_exception

    def _run_validation(self, test_results_storage: Dict, validator: Optional[Validator]):
        if not self.action.with_validation:
            return
        if validator is None:
            logger.debug(
                "The validator is None -- the validation should be skipped, "
                "most probably this test stage was run from a dependency chain"
            )
            return

        validator.validate(self.stage_results, test_results_storage)

    def run_once(
        self,
        data_collector: DataCollector,
        test_results_storage: OrderedDict,
        validator: Optional[Validator],
    ):
        logger.info(f'Begin stage "{self.name}"')
        assert isinstance(test_results_storage, OrderedDict)
        logger.debug(f'For test stage "{self.name}": test_results_storage.keys = {list(test_results_storage.keys())}')

        for dep_stage in self.get_depends_stages():
            # Processing all dependency stages of the current test.
            # Note that
            # * the stages may run their own dependency stages -- they will compose so called "dependency chain"
            # * the dependency stages are run with `validator = None`
            #   to avoid validation of stages that are run from the dependency chain.
            logger.debug(f'For test stage "{self.name}": Before running dep. stage "{dep_stage.name}"')
            dep_stage.run_once(data_collector, test_results_storage, validator=None)
            logger.debug(f'For test stage "{self.name}": After running dep. stage "{dep_stage.name}"')

        if self.was_processed:
            self._reraise_stage_exception_if_was_failed()
            # if we are here, then the stage was processed without exceptions
            logger.info(f"The stage {self.name} was already processed SUCCESSFULLY")

            # Run validation here for the rare case if this test now is being run *not* from a dependency chain
            # (i.e. the test is run with `validator != None`),
            # but the test already has been run from some dependency chain earlier.
            self._run_validation(test_results_storage, validator)

            logger.info(f'End stage "{self.name}"')
            return

        if self.name in test_results_storage:
            raise RuntimeError(
                f'Error: For test stage "{self.name}": '
                f"another OTXTestStage with name {self.name} has been run already"
            )

        try:
            logger.info(f'For test stage "{self.name}": Before running main action')
            self.stage_results = self.action(data_collector=data_collector, results_prev_stages=test_results_storage)
            logger.info(f'For test stage "{self.name}": After running main action')
            self.was_processed = True
            test_results_storage[self.name] = self.stage_results
            logger.debug(
                f'For test stage "{self.name}": after addition test_results_storage.keys = '
                f"{list(test_results_storage.keys())}"
            )
        except Exception as e:
            logger.info(
                f'For test stage "{self.name}": After running action for stage {self.name} -- CAUGHT EXCEPTION:\n{e}'
            )
            logger.info(f'End stage "{self.name}"')
            self.stored_exception = e
            self.was_processed = True
            raise e

        # The validation step is made outside the central try...except clause, since if the test was successful, but
        # the quality numbers were lower than expected, the result of the stage still may be re-used
        # in other stages.
        self._run_validation(test_results_storage, validator)
        logger.info(f'End stage "{self.name}"')



"""Model configuration of EfficientNetB2B-MaskRCNN model for Rotated-Detection Task."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name

_base_ = [
    "../../../../common/adapters/mmcv/configs/backbones/efficientnet_b2b.yaml",
    "../../base/models/detector.py",
]

task = "instance-segmentation"

model = dict(
    type="CustomMaskRCNN",  # Use CustomMaskRCNN for Incremental Learning
    neck=dict(type="FPN", in_channels=[24, 48, 120, 352], out_channels=80, num_outs=5),
    rpn_head=dict(
        type="RPNHead",
        in_channels=80,
        feat_channels=80,
        anchor_generator=dict(type="AnchorGenerator", scales=[8], ratios=[0.5, 1.0, 2.0], strides=[4, 8, 16, 32, 64]),
        bbox_coder=dict(type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[1.0, 1.0, 1.0, 1.0]),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="L1Loss", loss_weight=1.0),
    ),
    roi_head=dict(
        type="CustomRoIHead",  # Use CustomROIHead for Ignore mode
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=0),
            out_channels=80,
            featmap_strides=[4, 8, 16, 32],
        ),
        bbox_head=dict(
            type="Shared2FCBBoxHead",
            in_channels=80,
            fc_out_channels=1024,
            roi_feat_size=7,
            num_classes=80,
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[0.1, 0.1, 0.2, 0.2]
            ),
            reg_class_agnostic=False,
            loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
            loss_bbox=dict(type="L1Loss", loss_weight=1.0),
        ),
        mask_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=14, sampling_ratio=0),
            out_channels=80,
            featmap_strides=[4, 8, 16, 32],
        ),
        mask_head=dict(
            type="FCNMaskHead",
            num_convs=4,
            in_channels=80,
            conv_out_channels=80,
            num_classes=80,
            loss_mask=dict(type="CrossEntropyLoss", use_mask=True, loss_weight=1.0),
        ),
    ),
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                match_low_quality=True,
                ignore_iof_thr=-1,
                gpu_assign_thr=300,
            ),
            sampler=dict(type="RandomSampler", num=256, pos_fraction=0.5, neg_pos_ub=-1, add_gt_as_proposals=False),
            allowed_border=-1,
            pos_weight=-1,
            debug=False,
        ),
        rpn_proposal=dict(
            nms_across_levels=False,
            nms_pre=2000,
            max_per_img=1000,
            nms=dict(type="nms", iou_threshold=0.8),
            min_bbox_size=0,
        ),
        rcnn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.5,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                match_low_quality=True,
                ignore_iof_thr=-1,
                gpu_assign_thr=300,
            ),
            sampler=dict(type="RandomSampler", num=256, pos_fraction=0.25, neg_pos_ub=-1, add_gt_as_proposals=True),
            mask_size=28,
            pos_weight=-1,
            debug=False,
        ),
    ),
    test_cfg=dict(
        rpn=dict(
            nms_across_levels=False,
            nms_pre=800,
            max_per_img=500,
            nms=dict(type="nms", iou_threshold=0.8),
            min_bbox_size=0,
        ),
        rcnn=dict(score_thr=0.05, nms=dict(type="nms", iou_threshold=0.7), max_per_img=500, mask_thr_binary=0.5),
    ),
)
load_from = "https://storage.openvinotoolkit.org/repositories/\
openvino_training_extensions/models/instance_segmentation/\
v2/efficientnet_b2b-mask_rcnn-576x576.pth"

evaluation = dict(interval=1, metric="mAP", save_best="mAP", iou_thr=[0.5])
fp16 = dict(loss_scale=512.0)
ignore = True



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import hashlib
import os
import shutil
import string
import sys
import time
from pathlib import Path

import requests
from openvino.model_zoo import _common, _reporting
from openvino.model_zoo._configuration import load_models
from openvino.model_zoo.download_engine.downloader import Downloader
from openvino.model_zoo.download_engine.postprocessing import PostprocUnpackArchive
from openvino.model_zoo.omz_converter import ModelOptimizerProperties, convert_to_onnx
from requests.exceptions import HTTPError

from otx.mpa.utils.file import MPA_CACHE

MPA_OMZ_CACHE = os.path.join(MPA_CACHE, "omz")
os.makedirs(MPA_OMZ_CACHE, exist_ok=True)


OMZ_PUBLIC_MODELS = dict(
    cls=[
        "alexnet",
        "caffenet",
        #  "convnext-tiny",                # omz_downloader does not support
        "densenet-121",
        "densenet-121-tf",
        "dla-34",
        "efficientnet-b0",
        "efficientnet-b0-pytorch",
        "efficientnet-v2-b0",
        "efficientnet-v2-s",
        "hbonet-1.0",
        "hbonet-0.25",
        "googlenet-v1",
        "googlenet-v1-tf",
        "googlenet-v2",
        "googlenet-v2-tf",
        "googlenet-v3",
        "googlenet-v3-pytorch",
        "googlenet-v4-tf",
        "inception-resnet-v2-tf",
        #  "levit-128s",                   # IR has hard-codeded batch size of 1
        "mixnet-l",
        "mobilenet-v1-0.25-128",
        "mobilenet-v1-1.0-224",
        "mobilenet-v1-1.0-224-tf",
        "mobilenet-v2",
        "mobilenet-v2-1.0-224",
        "mobilenet-v2-pytorch",
        "mobilenet-v2-1.4-224",
        "mobilenet-v3-small-1.0-224-tf",
        "mobilenet-v3-large-1.0-224-tf",
        #  "nfnet-f0",                     # mo 2022.2 bug
        #  "regnetx-3.2gf",                # omz_converter does not support
        "octave-resnet-26-0.25",
        #  "repvgg-a0",                    # trainig and inference architecture are difference
        #  "repvgg-b1",                    # trainig and inference architecture are difference
        #  "repvgg-b3",                    # trainig and inference architecture are difference
        #  "resnest-50-pytorch",           # IR has hard-coded batch size of 1
        "resnet-18-pytorch",
        "resnet-34-pytorch",
        "resnet-50-pytorch",
        "resnet-50-tf",
        #  "rexnet-v1-x1.0",               # IR has hard-coded batch size of 1
        "se-inception",
        "se-resnet-50",
        "se-resnext-50",
        "shufflenet-v2-x0.5",
        #  "shufflenet-v2-x1.0",           # IR has hard-coded batch size of 1
        "squeezenet1.0",
        "squeezenet1.1",
        #  "swin-tiny-patch4-window7-224", # IR has hard-coded batch size of 1
        #  "t2t-vit-14",                   # IR has hard-coded batch size of 1
        "vgg16",
        "vgg19",
    ],
    det=[],
    seg=[],
)


AVAILABLE_OMZ_MODELS = [model for models in OMZ_PUBLIC_MODELS.values() for model in models]


class NameSpace:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def _get_etag(url):
    try:
        response = requests.head(url, allow_redirects=True)
        if response.status_code != 200:
            return None
        else:
            return response.headers.get("ETag", None)
    except HTTPError as e:
        return None


def _get_ir_path(directory):
    directory = Path(directory)
    model_path = list(directory.glob("**/*.xml"))
    weight_path = list(directory.glob("**/*.bin"))
    if model_path and weight_path:
        assert len(model_path) == 1 and len(weight_path) == 1
        return dict(model_path=model_path[0], weight_path=weight_path[0])
    else:
        return None


def _run_pre_convert(reporter, model, output_dir, args):
    script = _common.MODEL_ROOT / model.subdirectory_ori / "pre-convert.py"
    if not script.exists():
        return True

    reporter.print_section_heading(
        "{}Running pre-convert script for {}",
        "(DRY RUN) " if args.dry_run else "",
        model.name,
    )

    cmd = [
        str(args.python),
        "--",
        str(script),
        "--",
        str(args.download_dir / model.subdirectory),
        str(output_dir / model.subdirectory),
    ]

    reporter.print("Pre-convert command: {}", _common.command_string(cmd))
    reporter.print(flush=True)

    success = True if args.dry_run else reporter.job_context.subprocess(cmd)
    reporter.print()

    return success


def _update_model(model):
    m = hashlib.sha256()
    for file in model.files:
        url = file.source.url
        etag = _get_etag(url)
        if etag is not None:
            m.update(bytes(etag, "utf-8"))
    model.subdirectory_ori = model.subdirectory
    model.subdirectory = Path(m.hexdigest())

    # FIXME: a bug from openvino-dev==2022.3.0
    # It has been fixed on master branch.
    # After upgrading openvino-dev, we can remove this temporary patch
    if getattr(model, "conversion_to_onnx_args") and not [
        arg for arg in model.conversion_to_onnx_args if arg.startswith("--model-path")
    ]:
        model.conversion_to_onnx_args.append("--model-path=")


def get_model_configuration(model_name):
    model_configurations = load_models(_common.MODEL_ROOT, {})
    for i, model in enumerate(model_configurations):
        if model.name == model_name:
            _update_model(model)
            return model
    return None


def download_model(model, download_dir=MPA_OMZ_CACHE, precisions={"FP32"}, force=False):
    download_dir = Path("") if download_dir is None else Path(download_dir)

    # TODO: need delicate cache management
    if not force and (download_dir / model.subdirectory).exists():
        target_file_names = []
        for postprocessing in model.postprocessing:
            if isinstance(postprocessing, PostprocUnpackArchive):
                target_file_names.append(postprocessing.file)

        done = [False for _ in model.files]
        for i, file in enumerate(model.files):
            filename = file.name
            if filename in target_file_names:
                # TODO
                # here, we assume unarchive is done
                done[i] = True
                continue
            if os.path.exists(download_dir / model.subdirectory / filename):
                done[i] = True

        if all(done):
            return

    reporter = Downloader.make_reporter("text")
    downloader = Downloader(precisions, download_dir)
    failed_models = downloader.bulk_download_model([model], reporter, 1, "text")
    if failed_models:
        reporter.print("FAILED:")
        for failed_model_name in failed_models:
            reporter.print(failed_model_name)
        sys.exit(1)


def convert_model(
    model,
    download_dir=MPA_OMZ_CACHE,
    output_dir=MPA_OMZ_CACHE,
    precisions={"FP32"},
    force=False,
):
    download_dir = Path("") if download_dir is None else Path(download_dir)
    output_dir = Path("") if output_dir is None else Path(output_dir)

    out = _get_ir_path(output_dir / model.subdirectory)
    if out and not force:
        return out

    namespace = NameSpace(
        python=shutil.which("python"),
        dry_run=False,
        download_dir=download_dir,
    )

    mo_executable = shutil.which("mo")

    if mo_executable:
        mo_path = Path(mo_executable)
    else:
        try:
            mo_path = Path(os.environ["INTEL_OPENVINO_DIR"]) / "tools/mo/openvino/tools/mo/mo.py"
            if not mo_path.exists():
                mo_path = Path(os.environ["INTEL_OPENVINO_DIR"]) / "tools/model_optimizer/mo.py"
        except KeyError:
            sys.exit(
                "Unable to locate Model Optimizer. "
                + "Use --mo or run setupvars.sh/setupvars.bat from the OpenVINO toolkit."
            )

    mo_path = mo_path.resolve()
    mo_cmd_prefix = [namespace.python, "--", str(mo_path)]

    if str(mo_path).lower().endswith(".py"):
        mo_dir = mo_path.parent
    else:
        mo_package_path, stderr = _common.get_package_path(namespace.python, "openvino.tools.mo")
        mo_dir = mo_package_path

        if mo_package_path is None:
            mo_package_path, stderr = _common.get_package_path(args.python, "mo")
            if mo_package_path is None:
                sys.exit("Unable to load Model Optimizer. Errors occurred: {}".format(stderr))
            mo_dir = mo_package_path.parent

    reporter = _reporting.Reporter(_reporting.DirectOutputContext())
    mo_props = ModelOptimizerProperties(
        cmd_prefix=mo_cmd_prefix,
        extra_args=[],
        base_dir=mo_dir,
    )
    shared_convert_args = (output_dir, namespace, mo_props, precisions)

    def convert(reporter, model, output_dir, namespace, mo_props, requested_precisions):
        if model.mo_args is None:
            reporter.print_section_heading("Skipping {} (no conversions defined)", model.name)
            reporter.print()
            return True

        model_precisions = requested_precisions & model.precisions
        if not model_precisions:
            reporter.print_section_heading("Skipping {} (all conversions skipped)", model.name)
            reporter.print()
            return True

        (output_dir / model.subdirectory).mkdir(parents=True, exist_ok=True)

        if not _run_pre_convert(reporter, model, output_dir, namespace):
            return False

        model_format = model.framework
        mo_extension_dir = mo_props.base_dir / "extensions"
        if not mo_extension_dir.exists():
            mo_extension_dir = mo_props.base_dir

        template_variables = {
            "config_dir": _common.MODEL_ROOT / model.subdirectory_ori,
            "conv_dir": output_dir / model.subdirectory,
            "dl_dir": namespace.download_dir / model.subdirectory,
            "mo_dir": mo_props.base_dir,
            "mo_ext_dir": mo_extension_dir,
        }

        if model.conversion_to_onnx_args:
            if not convert_to_onnx(reporter, model, output_dir, namespace, template_variables):
                return False
            model_format = "onnx"

        expanded_mo_args = [string.Template(arg).substitute(template_variables) for arg in model.mo_args]

        for model_precision in sorted(model_precisions):
            data_type = model_precision.split("-")[0]
            layout_string = ",".join(
                "{}({})".format(input.name, input.layout) for input in model.input_info if input.layout
            )
            shape_string = ",".join(str(input.shape) for input in model.input_info if input.shape)

            if layout_string:
                expanded_mo_args.append("--layout={}".format(layout_string))
            if shape_string:
                expanded_mo_args.append("--input_shape={}".format(shape_string))

            mo_cmd = [
                *mo_props.cmd_prefix,
                "--framework={}".format(model_format),
                "--data_type={}".format(data_type),
                "--output_dir={}".format(output_dir / model.subdirectory / model_precision),
                "--model_name={}".format(model.name),
                "--input={}".format(",".join(input.name for input in model.input_info)),
                *expanded_mo_args,
                *mo_props.extra_args,
            ]

            reporter.print_section_heading(
                "{}Converting {} to IR ({})",
                "(DRY RUN) " if namespace.dry_run else "",
                model.name,
                model_precision,
            )

            reporter.print("Conversion command: {}", _common.command_string(mo_cmd))

            if not namespace.dry_run:
                reporter.print(flush=True)

                if not reporter.job_context.subprocess(mo_cmd):
                    # NOTE: mo returns non zero return code (245) even though it successfully generate IR
                    cur_time = time.time()
                    time_threshold = 5
                    xml_path = output_dir / model.subdirectory / model_precision / f"{model.name}.xml"
                    bin_path = output_dir / model.subdirectory / model_precision / f"{model.name}.bin"
                    if not (
                        os.path.exists(xml_path)
                        and os.path.exists(bin_path)
                        and os.path.getmtime(xml_path) - cur_time < time_threshold
                        and os.path.getmtime(bin_path) - cur_time < time_threshold
                    ):
                        return False

            reporter.print()

        return True

    results = []
    models = []
    if model.model_stages:
        for model_stage in model.model_stages:
            results.append(convert(reporter, model_stage, *shared_convert_args))
            models.append(model_stage)
    else:
        results.append(convert(reporter, model, *shared_convert_args))
        models.append(model)

    failed_models = [model.name for model, successful in zip(models, results) if not successful]

    if failed_models:
        reporter.print("FAILED:")
        for failed_model_name in failed_models:
            reporter.print(failed_model_name)
        sys.exit(1)

    return _get_ir_path(output_dir / model.subdirectory)


def get_omz_model(model_name, download_dir=MPA_OMZ_CACHE, output_dir=MPA_OMZ_CACHE, force=False):
    model = get_model_configuration(model_name)
    download_model(model, download_dir=download_dir, force=force)
    return convert_model(model, download_dir=download_dir, output_dir=output_dir, force=force)



"""Openvino Task of OTX Classification."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import logging
import os
import tempfile
import warnings
from typing import Any, Dict, Optional, Tuple, Union
from zipfile import ZipFile

import numpy as np
from addict import Dict as ADDict
from compression.api import DataLoader
from compression.engines.ie_engine import IEEngine
from compression.graph import load_model, save_model
from compression.graph.model_utils import compress_model_weights, get_nodes_by_type
from compression.pipeline.initializer import create_pipeline

from otx.algorithms.classification.adapters.openvino import model_wrappers
from otx.algorithms.classification.configs import ClassificationConfig
from otx.algorithms.classification.utils import get_multihead_class_info
from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import (
    InferenceParameters,
    default_progress_callback,
)
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.metadata import FloatMetadata, FloatType
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.serialization.label_mapper import LabelSchemaMapper, label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.exportable_code import demo
from otx.api.usecases.exportable_code.inference import BaseInferencer
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    ClassificationToAnnotationConverter,
)
from otx.api.usecases.tasks.interfaces.deployment_interface import IDeploymentTask
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.explain_interface import IExplainTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.dataset_utils import add_saliency_maps_to_dataset_item

try:
    from openvino.model_zoo.model_api.adapters import OpenvinoAdapter, create_core
    from openvino.model_zoo.model_api.models import Model
except ImportError:
    warnings.warn("ModelAPI was not found.")

logger = logging.getLogger(__name__)


# TODO: refactoring to Sphinx style.
class ClassificationOpenVINOInferencer(BaseInferencer):
    """ClassificationOpenVINOInferencer class in OpenVINO task."""

    @check_input_parameters_type()
    def __init__(
        self,
        hparams: ClassificationConfig,
        label_schema: LabelSchemaEntity,
        model_file: Union[str, bytes],
        weight_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):  # pylint: disable=unused-argument
        """Inferencer implementation for OTXDetection using OpenVINO backend.

        :param model: Path to model to load, `.xml`, `.bin` or `.onnx` file.
        :param hparams: Hyper parameters that the model should use.
        :param num_requests: Maximum number of requests that the inferencer can make.
            Good value is the number of available cores. Defaults to 1.
        :param device: Device to run inference on, such as CPU, GPU or MYRIAD. Defaults to "CPU".
        """

        multilabel = len(label_schema.get_groups(False)) > 1 and len(label_schema.get_groups(False)) == len(
            label_schema.get_labels(include_empty=False)
        )
        hierarchical = not multilabel and len(label_schema.get_groups(False)) > 1
        multihead_class_info = {}
        if hierarchical:
            multihead_class_info = get_multihead_class_info(label_schema)

        self.label_schema = label_schema

        model_adapter = OpenvinoAdapter(
            create_core(), model_file, weight_file, device=device, max_num_requests=num_requests
        )
        self.configuration = {
            "multilabel": multilabel,
            "hierarchical": hierarchical,
            "multihead_class_info": multihead_class_info,
        }
        self.model = Model.create_model("otx_classification", model_adapter, self.configuration, preload=True)

        self.converter = ClassificationToAnnotationConverter(self.label_schema)

    @check_input_parameters_type()
    def pre_process(self, image: np.ndarray) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
        """Pre-process function of OpenVINO Classification Inferencer."""

        return self.model.preprocess(image)

    @check_input_parameters_type()
    def post_process(
        self, prediction: Dict[str, np.ndarray], metadata: Dict[str, Any]
    ) -> Optional[AnnotationSceneEntity]:
        """Post-process function of OpenVINO Classification Inferencer."""

        classification = self.model.postprocess(prediction, metadata)
        return self.converter.convert_to_annotation(classification, metadata)

    @check_input_parameters_type()
    def predict(self, image: np.ndarray) -> Tuple[AnnotationSceneEntity, np.ndarray, np.ndarray, np.ndarray, Any]:
        """Predict function of OpenVINO Classification Inferencer."""

        image, metadata = self.pre_process(image)
        raw_predictions = self.forward(image)
        predictions = self.post_process(raw_predictions, metadata)
        probs, actmap, repr_vectors, act_score = self.model.postprocess_aux_outputs(raw_predictions, metadata)

        return predictions, probs, actmap, repr_vectors, act_score

    @check_input_parameters_type()
    def forward(self, image: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """Forward function of OpenVINO Classification Inferencer."""

        return self.model.infer_sync(image)


class OTXOpenVinoDataLoader(DataLoader):
    """DataLoader implementation for ClassificationOpenVINOTask."""

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def __init__(self, dataset: DatasetEntity, inferencer: BaseInferencer):
        super().__init__(config=None)
        self.dataset = dataset
        self.inferencer = inferencer

    @check_input_parameters_type()
    def __getitem__(self, index: int):
        """Get item from dataset."""

        image = self.dataset[index].numpy
        annotation = self.dataset[index].annotation_scene
        inputs, metadata = self.inferencer.pre_process(image)

        return (index, annotation), inputs, metadata

    def __len__(self):
        """Get length of dataset."""

        return len(self.dataset)


class ClassificationOpenVINOTask(IDeploymentTask, IInferenceTask, IEvaluationTask, IExplainTask, IOptimizationTask):
    """Task implementation for OTXClassification using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment):
        self.task_environment = task_environment
        self.hparams = self.task_environment.get_hyper_parameters(ClassificationConfig)
        self.model = self.task_environment.model
        self.inferencer = self.load_inferencer()

    def load_inferencer(self) -> ClassificationOpenVINOInferencer:
        """load_inferencer function of ClassificationOpenVINOTask."""

        if self.model is None:
            raise RuntimeError("load_inferencer failed, model is None")

        return ClassificationOpenVINOInferencer(
            self.hparams,
            self.task_environment.label_schema,
            self.model.get_data("openvino.xml"),
            self.model.get_data("openvino.bin"),
        )

    # pylint: disable-msg=too-many-locals
    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self, dataset: DatasetEntity, inference_parameters: Optional[InferenceParameters] = None
    ) -> DatasetEntity:
        """Infer function of ClassificationOpenVINOTask."""

        update_progress_callback = default_progress_callback
        dump_features = False
        process_saliency_maps = False
        explain_predicted_classes = True
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore
            dump_features = not inference_parameters.is_evaluation
            process_saliency_maps = inference_parameters.process_saliency_maps
            explain_predicted_classes = inference_parameters.explain_predicted_classes

        dataset_size = len(dataset)
        for i, dataset_item in enumerate(dataset, 1):
            predicted_scene, probs, saliency_map, repr_vector, act_score = self.inferencer.predict(dataset_item.numpy)
            item_labels = predicted_scene.annotations[0].get_labels()
            dataset_item.append_labels(item_labels)
            active_score_media = FloatMetadata(name="active_score", value=act_score, float_type=FloatType.ACTIVE_SCORE)
            dataset_item.append_metadata_item(active_score_media, model=self.model)

            probs_meta = TensorEntity(name="probabilities", numpy=probs.reshape(-1))
            dataset_item.append_metadata_item(probs_meta, model=self.model)

            if dump_features:
                if saliency_map is not None and repr_vector is not None:
                    feature_vec_media = TensorEntity(name="representation_vector", numpy=repr_vector.reshape(-1))
                    dataset_item.append_metadata_item(feature_vec_media, model=self.model)

                    add_saliency_maps_to_dataset_item(
                        dataset_item=dataset_item,
                        saliency_map=saliency_map,
                        model=self.model,
                        labels=self.task_environment.get_labels(),
                        predicted_scored_labels=item_labels,
                        explain_predicted_classes=explain_predicted_classes,
                        process_saliency_maps=process_saliency_maps,
                    )
                else:
                    warnings.warn(
                        "Could not find Feature Vector and Saliency Map in OpenVINO output. "
                        "Please rerun OpenVINO export or retrain the model."
                    )
            update_progress_callback(int(i / dataset_size * 100))
        return dataset

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Explain function of ClassificationOpenVINOTask."""

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if explain_parameters is not None:
            update_progress_callback = explain_parameters.update_progress  # type: ignore
            process_saliency_maps = explain_parameters.process_saliency_maps
            explain_predicted_classes = explain_parameters.explain_predicted_classes

        dataset_size = len(dataset)
        for i, dataset_item in enumerate(dataset, 1):
            predicted_scene, _, saliency_map, _, _ = self.inferencer.predict(dataset_item.numpy)
            item_labels = predicted_scene.annotations[0].get_labels()
            dataset_item.append_labels(item_labels)
            add_saliency_maps_to_dataset_item(
                dataset_item=dataset_item,
                saliency_map=np.copy(saliency_map),
                model=self.model,
                labels=self.task_environment.get_labels(),
                predicted_scored_labels=item_labels,
                explain_predicted_classes=explain_predicted_classes,
                process_saliency_maps=process_saliency_maps,
            )
            update_progress_callback(int(i / dataset_size * 100))
        return dataset

    @check_input_parameters_type()
    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None):
        """Evaluate function of ClassificationOpenVINOTask."""

        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric," "but parameter is ignored. Use accuracy instead."
            )
        output_resultset.performance = MetricsHelper.compute_accuracy(output_resultset).get_performance()

    @check_input_parameters_type()
    def deploy(self, output_model: ModelEntity) -> None:
        """Deploy function of ClassificationOpenVINOTask."""

        logger.info("Deploying the model")

        work_dir = os.path.dirname(demo.__file__)
        parameters = {}  # type: Dict[Any, Any]
        parameters["type_of_model"] = "otx_classification"
        parameters["converter_type"] = "CLASSIFICATION"
        parameters["model_parameters"] = self.inferencer.configuration
        parameters["model_parameters"]["labels"] = LabelSchemaMapper.forward(self.task_environment.label_schema)

        if self.model is None:
            raise RuntimeError("deploy failed, model is None")

        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, "w") as arch:
            # model files
            arch.writestr(os.path.join("model", "model.xml"), self.model.get_data("openvino.xml"))
            arch.writestr(os.path.join("model", "model.bin"), self.model.get_data("openvino.bin"))
            arch.writestr(os.path.join("model", "config.json"), json.dumps(parameters, ensure_ascii=False, indent=4))
            # model_wrappers files
            for root, _, files in os.walk(os.path.dirname(model_wrappers.__file__)):
                for file in files:
                    file_path = os.path.join(root, file)
                    arch.write(
                        file_path, os.path.join("python", "model_wrappers", file_path.split("model_wrappers/")[1])
                    )
            # other python files
            arch.write(os.path.join(work_dir, "requirements.txt"), os.path.join("python", "requirements.txt"))
            arch.write(os.path.join(work_dir, "LICENSE"), os.path.join("python", "LICENSE"))
            arch.write(os.path.join(work_dir, "README.md"), os.path.join("python", "README.md"))
            arch.write(os.path.join(work_dir, "demo.py"), os.path.join("python", "demo.py"))
        output_model.exportable_code = zip_buffer.getvalue()
        logger.info("Deploying completed")

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):  # pylint: disable=too-many-locals
        """Optimize function of ClassificationOpenVINOTask."""

        if optimization_type is not OptimizationType.POT:
            raise ValueError("POT is the only supported optimization type for OpenVino models")

        dataset = dataset.get_subset(Subset.TRAINING)
        data_loader = OTXOpenVinoDataLoader(dataset, self.inferencer)

        if self.model is None:
            raise RuntimeError("optimize failed, model is None")

        with tempfile.TemporaryDirectory() as tempdir:
            xml_path = os.path.join(tempdir, "model.xml")
            bin_path = os.path.join(tempdir, "model.bin")
            with open(xml_path, "wb") as f:
                f.write(self.model.get_data("openvino.xml"))
            with open(bin_path, "wb") as f:
                f.write(self.model.get_data("openvino.bin"))

            model_config = ADDict({"model_name": "openvino_model", "model": xml_path, "weights": bin_path})

            model = load_model(model_config)

            if get_nodes_by_type(model, ["FakeQuantize"]):
                raise RuntimeError("Model is already optimized by POT")

        if optimization_parameters is not None:
            optimization_parameters.update_progress(10, None)

        engine_config = ADDict({"device": "CPU"})

        stat_subset_size = self.hparams.pot_parameters.stat_subset_size
        preset = self.hparams.pot_parameters.preset.name.lower()

        algorithms = [
            {
                "name": "DefaultQuantization",
                "params": {
                    "target_device": "ANY",
                    "preset": preset,
                    "stat_subset_size": min(stat_subset_size, len(data_loader)),
                    "shuffle_data": True,
                },
            }
        ]

        engine = IEEngine(config=engine_config, data_loader=data_loader, metric=None)

        pipeline = create_pipeline(algorithms, engine)

        compressed_model = pipeline.run(model)

        compress_model_weights(compressed_model)

        if optimization_parameters is not None:
            optimization_parameters.update_progress(90, None)

        with tempfile.TemporaryDirectory() as tempdir:
            save_model(compressed_model, tempdir, model_name="model")
            with open(os.path.join(tempdir, "model.xml"), "rb") as f:
                output_model.set_data("openvino.xml", f.read())
            with open(os.path.join(tempdir, "model.bin"), "rb") as f:
                output_model.set_data("openvino.bin", f.read())

        output_model.set_data("label_schema.json", label_schema_to_bytes(self.task_environment.label_schema))

        # set model attributes for quantized model
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.POT
        output_model.optimization_methods = [OptimizationMethod.QUANTIZATION]
        output_model.precision = [ModelPrecision.INT8]

        self.model = output_model
        self.inferencer = self.load_inferencer()

        if optimization_parameters is not None:
            optimization_parameters.update_progress(100, None)
        logger.info("POT optimization completed")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Optional

from ..registry import Registry


class OperationRegistry(Registry):
    def __init__(self, name, add_name_as_attr=False):
        super().__init__(name, add_name_as_attr)
        self._registry_dict_by_type = {}

    def register(self, name: Optional[Any] = None):
        def wrap(obj):
            layer_name = name
            if layer_name is None:
                layer_name = obj.__name__
            layer_type = obj.TYPE
            layer_version = obj.VERSION
            assert layer_type != "" and layer_version >= 0
            if self._add_name_as_attr:
                setattr(obj, self.REGISTERED_NAME_ATTR, layer_name)
            self._register(obj, layer_name, layer_type, layer_version)
            return obj

        return wrap

    def _register(self, obj, name, type, version):
        super()._register(obj, name)
        if type not in self._registry_dict_by_type:
            self._registry_dict_by_type[type] = {}
        if version in self._registry_dict_by_type[type]:
            raise KeyError(f"{version} is already registered in {type}")
        self._registry_dict_by_type[type][version] = obj

    def get_by_name(self, name):
        return self.get(name)

    def get_by_type_version(self, type, version):
        if type not in self._registry_dict_by_type:
            raise KeyError(f"type {type} is not registered in {self._name}")
        if version not in self._registry_dict_by_type[type]:
            raise KeyError(f"version {version} is not registered in {type} of {self._name}")
        return self._registry_dict_by_type[type][version]


OPS = OperationRegistry("ov ops")



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import pytest
from mmcv.utils import Config

from otx.algorithms.common.tasks import BaseTask
from otx.algorithms.common.tasks.nncf_base import NNCFBaseTask
from otx.algorithms.segmentation.tasks import SegmentationNNCFTask
from otx.api.configuration.helper import create
from otx.api.entities.metrics import NullPerformance
from otx.api.entities.model_template import parse_model_template
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_SEG_TEMPLATE_DIR,
    generate_otx_dataset,
    init_environment,
)


class TestOTXSegTaskNNCF:
    @pytest.fixture(autouse=True)
    def setup(self, otx_model, tmp_dir_path) -> None:
        model_template = parse_model_template(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        task_env = init_environment(hyper_parameters, model_template)
        self.model = otx_model
        self.seg_nncf_task = SegmentationNNCFTask(task_env, output_path=str(tmp_dir_path))

    @e2e_pytest_unit
    def test_save_model(self, mocker):
        mocker.patch("torch.load", return_value="")
        self.seg_nncf_task._recipe_cfg = Config({})
        self.seg_nncf_task._model_cfg = Config({"model": {}})
        self.seg_nncf_task.save_model(self.model)

        assert self.model.get_data("weights.pth")
        assert self.model.get_data("label_schema.json")

    @e2e_pytest_unit
    def test_optimize(self, mocker):
        from otx.algorithms.common.adapters.mmcv.hooks import OTXLoggerHook

        self.dataset = generate_otx_dataset()

        mock_lcurve_val = OTXLoggerHook.Curve()
        mock_lcurve_val.x = [0, 1]
        mock_lcurve_val.y = [0.1, 0.2]

        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value={"final_ckpt": ""})
        self.seg_nncf_task._learning_curves = {f"val/{self.seg_nncf_task.metric}": mock_lcurve_val}
        mocker.patch.object(SegmentationNNCFTask, "save_model")
        self.seg_nncf_task.optimize(OptimizationType.NNCF, self.dataset, self.model)

        mock_run_task.assert_called_once()
        assert self.model.performance != NullPerformance()
        assert self.model.performance.score.value == 0.2

    @e2e_pytest_unit
    def test_initialize(self, mocker):
        """Test initialize method in OTXDetTaskNNCF."""
        options = {}
        self.seg_nncf_task._initialize(options)

        assert "model_builder" in options
        assert NNCFBaseTask.model_builder == options["model_builder"].func



import unittest
from typing import List

import numpy as np
import torch
from mmcv import ConfigDict
from mmdet.datasets import build_dataloader, build_dataset

from otx.algorithms.detection.adapters.mmdet.data import (  # noqa: F401
    ImageTilingDataset,
)
from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.utils.shape_factory import ShapeFactory
from tests.test_helpers import generate_random_annotated_image
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def create_otx_dataset(height: int, width: int, labels: List[str]):
    """Create a random OTX dataset

    Args:
        height (int): The height of the image
        width (int): The width of the image

    Returns:
        DatasetEntity: OTX dataset entity
        List[LabelEntity]: The list of labels
    """
    labels = []
    for label in ["rectangle", "ellipse", "triangle"]:
        labels.append(LabelEntity(name=label, domain=Domain.DETECTION))
    image, anno_list = generate_random_annotated_image(width, height, labels)
    image = Image(data=image)
    annotation_scene = AnnotationSceneEntity(annotations=anno_list, kind=AnnotationSceneKind.ANNOTATION)
    dataset_item = DatasetItemEntity(media=image, annotation_scene=annotation_scene)
    return DatasetEntity([dataset_item]), labels


class TestTilingDetection(unittest.TestCase):
    """Test the tiling functionality"""

    def setUp(self) -> None:
        """Setup the test case"""
        self.height = 1024
        self.width = 1024
        self.label_names = ["rectangle", "ellipse", "triangle"]
        self.tile_cfg = dict(
            tile_size=np.random.randint(low=100, high=500),
            overlap_ratio=np.random.uniform(low=0.0, high=0.5),
            max_per_img=np.random.randint(low=1, high=10000),
        )
        self.dataloader_cfg = dict(samples_per_gpu=1, workers_per_gpu=1)
        self.otx_dataset, self.labels = create_otx_dataset(self.height, self.width, self.label_names)

        img_norm_cfg = dict(mean=[0, 0, 0], std=[255, 255, 255], to_rgb=True)

        self.train_data_cfg = ConfigDict(
            dict(
                type="ImageTilingDataset",
                pipeline=[
                    dict(type="Resize", img_scale=(self.height, self.width), keep_ratio=False),
                    dict(type="RandomFlip", flip_ratio=0.5),
                    dict(type="Pad", size_divisor=32),
                    dict(type="DefaultFormatBundle"),
                    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels"]),
                ],
                dataset=dict(
                    type="OTXDetDataset",
                    pipeline=[
                        dict(type="LoadImageFromOTXDataset"),
                        dict(
                            type="LoadAnnotationFromOTXDataset",
                            with_bbox=True,
                            with_mask=False,
                            domain=Domain.DETECTION,
                            min_size=-1,
                        ),
                    ],
                    otx_dataset=self.otx_dataset,
                    labels=self.labels,
                    domain=Domain.DETECTION,
                ),
                **self.tile_cfg
            )
        )

        self.test_data_cfg = ConfigDict(
            dict(
                type="ImageTilingDataset",
                pipeline=[
                    dict(
                        type="MultiScaleFlipAug",
                        img_scale=(self.height, self.width),
                        flip=False,
                        transforms=[
                            dict(type="Resize", keep_ratio=False),
                            dict(type="Normalize", **img_norm_cfg),
                            dict(type="ImageToTensor", keys=["img"]),
                            dict(type="Collect", keys=["img"]),
                        ],
                    )
                ],
                dataset=dict(
                    type="OTXDetDataset",
                    pipeline=[dict(type="LoadImageFromOTXDataset")],
                    otx_dataset=self.otx_dataset.with_empty_annotations(),
                    labels=list(self.labels),
                    domain=Domain.DETECTION,
                ),
                test_mode=True,
                **self.tile_cfg
            )
        )

    @e2e_pytest_unit
    def test_tiling_train_dataloader(self):
        """Test that the training dataloader is built correctly for tiling"""

        dataset = build_dataset(self.train_data_cfg)
        train_dataloader = build_dataloader(dataset, **self.dataloader_cfg)
        for data in train_dataloader:
            self.assertIsInstance(data["img"].data[0], torch.Tensor)
            self.assertIsInstance(data["gt_bboxes"].data[0][0], torch.Tensor)
            self.assertIsInstance(data["gt_labels"].data[0][0], torch.Tensor)

    @e2e_pytest_unit
    def test_tiling_test_dataloader(self):
        """Test that the testing dataloader is built correctly for tiling"""

        dataset = build_dataset(self.test_data_cfg)
        stride = int((1 - self.tile_cfg["overlap_ratio"]) * self.tile_cfg["tile_size"])
        num_tile_rows = ((self.height - self.tile_cfg["tile_size"]) // stride) + 1
        num_tile_cols = ((self.width - self.tile_cfg["tile_size"]) // stride) + 1
        # +1 for the original image
        self.assertEqual(len(dataset), (num_tile_rows * num_tile_cols) + 1, "Incorrect number of tiles")

        test_dataloader = build_dataloader(dataset, **self.dataloader_cfg)
        for data in test_dataloader:
            self.assertIsInstance(data["img"][0], torch.Tensor)
            self.assertNotIn("gt_bboxes", data)
            self.assertNotIn("gt_labels", data)

    @e2e_pytest_unit
    def test_inference_merge(self):
        """Test that the inference merge works correctly"""
        dataset = build_dataset(self.test_data_cfg)

        # create simulated inference results
        results: List[List[np.ndarray]] = []
        for i in range(len(dataset)):
            results.append([])
            for _ in range(len(self.labels)):
                results[i].append(np.zeros((0, 5), dtype=np.float32))

        # generate tile predictions
        for i in range(len(dataset)):
            img_width, img_height = self.tile_cfg["tile_size"], self.tile_cfg["tile_size"]
            if i == 0:
                # first index belongs is the full image
                img_width, img_height = self.width, self.height

            _, anno_list = generate_random_annotated_image(img_width, img_height, self.labels)
            for anno in anno_list:
                shape = ShapeFactory.shape_as_rectangle(anno.shape)
                bbox = np.array([shape.x1, shape.y1, shape.x2, shape.y2], np.float32)
                bbox *= np.tile([img_width, img_height], 2)
                score_bbox = np.array([*bbox, np.random.rand()], np.float32)
                label_idx = self.label_names.index(anno.get_labels()[0].name)
                results[i][label_idx] = np.append(results[i][label_idx], [score_bbox], axis=0)

        merged_bbox_results = dataset.merge(results)
        self.assertEqual(len(merged_bbox_results), dataset.num_samples)
